# -*- coding: UTF-8 -*-
"""
Parser utilities test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest
from yokadi.ycli import parseutils

gTaskLineToParsedStructList = [
    ("project some text @keyword1 @keyword2=12 some other text", ("project", "some text some other text",
                                                                  {"keyword1": None, "keyword2": 12})),
    ("project ééé", ("project", "ééé", {})),
    ("project let's include quotes\"", ("project", "let's include quotes\"", {})),
    ("   project this  one has  extra spaces  ", ("project", "this one has extra spaces", {})),
]


class ParseUtilsTestCase(unittest.TestCase):
    def testExtractKeywords(self):
        for src, dst in gTaskLineToParsedStructList:
            result = parseutils.parseLine(src)
            self.assertEqual(result, dst)

    def testCreateLine(self):
        for dummy, parsedStruct in gTaskLineToParsedStructList:
            # We do not check the result of createLine() against the
            # original task line because there are many ways to write the same
            # taskLine.
            taskLine = parseutils.createLine(*parsedStruct)
            result = parseutils.parseLine(taskLine)
            self.assertEqual(result, parsedStruct)



# -*- coding: UTF-8 -*-
"""
Mass edit test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from yokadi.core import db
from yokadi.core.db import NOTE_KEYWORD
from yokadi.core import dbutils
from yokadi.core.yokadiexception import YokadiException
from yokadi.ycli import massedit
from yokadi.ycli import tui
from yokadi.ycli.massedit import MEditEntry, parseMEditText, ParseError


class MassEditTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        # FIXME: Do this in db
        dbutils.getOrCreateKeyword(NOTE_KEYWORD, interactive=False)
        self.session = db.getSession()
        tui.clearInputAnswers()

    def testApplyMEditChanges(self):
        prj = dbutils.getOrCreateProject("p1", interactive=False)
        t1 = dbutils.addTask("p1", "Change text", {})
        tui.addInputAnswers("y", "y")
        t2 = dbutils.addTask("p1", "Change keywords", {"k1": None, "k2": 1})
        t3 = dbutils.addTask("p1", "Done", {})
        t3.status = "started"
        t4 = dbutils.addTask("p1", "Deleted", {})
        t5 = dbutils.addTask("p1", "Moved", {})
        self.session.commit()
        deletedId = t4.id

        oldList = massedit.createEntriesForProject(prj)
        newList = [
            MEditEntry(None, "new", u"Added", {}),
            MEditEntry(t1.id, "new", u"New text", {}),
            MEditEntry(t2.id, "new", u"Change keywords", {"k2": 2, "k3": None}),
            MEditEntry(t5.id, "new", u"Moved", {}),
            MEditEntry(t3.id, "done", u"Done", {}),
        ]

        massedit.applyChanges(prj, oldList, newList, interactive=False)
        self.session.commit()

        newTask = self.session.query(db.Task).filter_by(title=u"Added").one()

        self.assertEqual(t1.title, u"New text")
        self.assertEqual(t2.getKeywordDict(), {"k2": 2, "k3": None})
        self.assertEqual(t3.status, "done")
        self.assertTrue(t3.doneDate)
        self.assertRaises(YokadiException, dbutils.getTaskFromId, deletedId)
        self.assertEqual(newTask.urgency, 5)
        self.assertEqual(t1.urgency, 4)
        self.assertEqual(t2.urgency, 3)
        self.assertEqual(t5.urgency, 2)
        self.assertEqual(t3.urgency, 1)

    def testApplyMEditChangesUnknownIds(self):
        prj = dbutils.getOrCreateProject("p1", interactive=False)
        t1 = dbutils.addTask("p1", "Foo", {})
        t2 = dbutils.addTask("p1", "Bar", {})

        oldList = massedit.createEntriesForProject(prj)
        newList = [
            MEditEntry(t1.id, "new", t1.title, {}),
            MEditEntry(t2.id + 1, "new", t2.title, {}),
        ]

        self.assertRaises(YokadiException, massedit.applyChanges, prj, oldList,
                          newList, interactive=False)

    def testParseMEditText(self):
        text = """1 N Hello
            4 N Some keywords @foo @bar=1
            6 S A started task
            12 D A done task
            - A newly added task
            - OneWordNewTask

            # A comment
            """
        expected = [
            MEditEntry(1, "new", u"Hello", {}),
            MEditEntry(4, "new", u"Some keywords", {"foo": None, "bar": 1}),
            MEditEntry(6, "started", u"A started task", {}),
            MEditEntry(12, "done", u"A done task", {}),
            MEditEntry(None, "new", u"A newly added task", {}),
            MEditEntry(None, "new", u"OneWordNewTask", {}),
        ]
        output = parseMEditText(text)

        self.assertEqual(output, expected)

    def testParseMEditTextErrors(self):
        testData = [
            # Duplicate id
            """
            1 N X
            1 N Y
            """,
            # Invalid id
            """
            A N X
            """,
            # Invalid status
            """
            1 z Y
            """,
            # Invalid line
            """
            bla
            """
        ]
        for text in testData:
            self.assertRaises(ParseError, parseMEditText, text)

    def testOnlyListTasks(self):
        prj = dbutils.getOrCreateProject("p1", interactive=False)
        dbutils.addTask("p1", "Task", {})
        dbutils.addTask("p1", "Note", {NOTE_KEYWORD: None})

        oldList = massedit.createEntriesForProject(prj)
        self.assertEqual(len(oldList), 1)

    def testCreateMEditText(self):
        e1 = MEditEntry(1, "N", "Hello", {})
        e2 = MEditEntry(2, "S", "Started", {})
        EXPECTED_TEXT = """1 N Hello
2 S Started

# doc1
#
# doc2
"""

        txt = massedit.createMEditText([e1, e2], docComment="doc1\n\ndoc2\n")
        self.assertEqual(txt, EXPECTED_TEXT)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Database access layer using SQL Alchemy

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""

import json
import os
from datetime import datetime
from uuid import uuid1

from sqlalchemy import create_engine, inspect
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import scoped_session, sessionmaker, relationship, declarative_base
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError
from sqlalchemy import Column, Integer, Boolean, Unicode, DateTime, Enum, ForeignKey, UniqueConstraint
from sqlalchemy.types import TypeDecorator, VARCHAR

from yokadi.core.recurrencerule import RecurrenceRule
from yokadi.core.yokadiexception import YokadiException

# Yokadi database version needed for this code
# If database config key DB_VERSION differs from this one a database migration
# is required
DB_VERSION = 12
DB_VERSION_KEY = "DB_VERSION"


class DbUserException(Exception):
    """
    This exception is for errors which are not caused by a failure in our code
    and which must be fixed by the user.
    """
    pass


Base = declarative_base()


NOTE_KEYWORD = "_note"


def uuidGenerator():
    return str(uuid1())


class Project(Base):
    __tablename__ = "project"
    id = Column(Integer, primary_key=True)
    uuid = Column(Unicode, unique=True, default=uuidGenerator, nullable=False)
    name = Column(Unicode, unique=True)
    active = Column(Boolean, default=True)
    tasks = relationship("Task", cascade="all", backref="project", cascade_backrefs=False)

    def __repr__(self):
        return self.name

    def merge(self, session, other):
        """Merge other into us

        This function calls session.commit() itself: we have to commit after
        moving the tasks but *before* deleting `other` otherwise when we delete
        `other` SQLAlchemy deletes its former tasks as well because it thinks
        they are still attached to `other`"""
        if self is other:
            raise YokadiException("Cannot merge a project into itself")

        for task in other.tasks:
            task.projectId = self.id

        session.commit()

        session.delete(other)
        session.commit()


class Keyword(Base):
    __tablename__ = "keyword"
    id = Column(Integer, primary_key=True)
    name = Column(Unicode, unique=True)
    tasks = association_proxy("taskKeywords", "task")
    taskKeywords = relationship("TaskKeyword", cascade="all", backref="keyword", cascade_backrefs=False)

    def __repr__(self):
        return self.name


class TaskKeyword(Base):
    __tablename__ = "task_keyword"
    __mapper_args__ = {"confirm_deleted_rows": False}
    id = Column(Integer, primary_key=True)
    taskId = Column("task_id", Integer, ForeignKey("task.id"), nullable=False)
    keywordId = Column("keyword_id", Integer, ForeignKey("keyword.id"), nullable=False)
    value = Column(Integer, default=None)

    __table_args__ = (
        UniqueConstraint("task_id", "keyword_id", name="task_keyword_uc"),
    )

    def __repr__(self):
        return "<TaskKeyword task={} keyword={} value={}>".format(self.task, self.keyword, self.value)


class RecurrenceRuleColumnType(TypeDecorator):
    """Represents an ydateutils.RecurrenceRule column
    """
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value:
            value = json.dumps(value.toDict())
        else:
            value = ""
        return value

    def process_result_value(self, value, dialect):
        if value:
            dct = json.loads(value)
            value = RecurrenceRule.fromDict(dct)
        else:
            value = RecurrenceRule()
        return value


class Task(Base):
    __tablename__ = "task"
    id = Column(Integer, primary_key=True)
    uuid = Column(Unicode, unique=True, default=uuidGenerator, nullable=False)
    title = Column(Unicode)
    creationDate = Column("creation_date", DateTime, nullable=False, default=datetime.now)
    dueDate = Column("due_date", DateTime, default=None)
    doneDate = Column("done_date", DateTime, default=None)
    description = Column(Unicode, default="", nullable=False)
    urgency = Column(Integer, default=0, nullable=False)
    status = Column(Enum("new", "started", "done"), default="new")
    recurrence = Column(RecurrenceRuleColumnType, nullable=False, default=RecurrenceRule())
    projectId = Column("project_id", Integer, ForeignKey("project.id"), nullable=False)
    taskKeywords = relationship("TaskKeyword", cascade="all", backref="task", cascade_backrefs=False)
    lock = relationship("TaskLock", cascade="all", backref="task", cascade_backrefs=False)

    def setKeywordDict(self, dct):
        """
        Defines keywords of a task.
        Dict is of the form: keywordName => value
        """
        session = getSession()
        for taskKeyword in self.taskKeywords:
            session.delete(taskKeyword)

        for name, value in list(dct.items()):
            keyword = session.query(Keyword).filter_by(name=name).one()
            session.add(TaskKeyword(task=self, keyword=keyword, value=value))

    def getKeywordDict(self):
        """
        Returns all keywords of a task as a dict of the form:
        keywordName => value
        """
        dct = {}
        for taskKeyword in self.taskKeywords:
            dct[taskKeyword.keyword.name] = taskKeyword.value
        return dct

    def getKeywordsAsString(self):
        """
        Returns all keywords as a string like "key1=value1, key2=value2..."
        """
        return ", ".join(list(("%s=%s" % k for k in list(self.getKeywordDict().items()))))

    def getUserKeywordsNameAsString(self):
        """
        Returns all keywords keys as a string like "@key1 @key2 @key3...".
        Internal keywords (starting with _) are ignored.
        """
        keywords = [k for k in list(self.getKeywordDict().keys()) if not k.startswith("_")]
        keywords.sort()
        if keywords:
            return " ".join(f"@{k}" for k in keywords)
        else:
            return ""

    def setStatus(self, status):
        """
        Defines the status of the task, taking care of updating the done date
        and doing the right thing for recurrent tasks
        """
        if self.recurrence and status == "done":
            self.dueDate = self.recurrence.getNext(self.dueDate)
        else:
            self.status = status
            if status == "done":
                self.doneDate = datetime.now().replace(second=0, microsecond=0)
            else:
                self.doneDate = None

    def setRecurrenceRule(self, rule):
        """Set recurrence and update the due date accordingly"""
        self.recurrence = rule
        self.dueDate = rule.getNext()

    @staticmethod
    def getNoteKeyword(session):
        return session.query(Keyword).filter_by(name=NOTE_KEYWORD).one()

    def toNote(self, session):
        session.add(TaskKeyword(task=self, keyword=Task.getNoteKeyword(session), value=None))
        try:
            session.flush()
        except IntegrityError:
            # Already a note
            session.rollback()
            return

    def toTask(self, session):
        noteKeyword = Task.getNoteKeyword(session)
        try:
            taskKeyword = session.query(TaskKeyword).filter_by(task=self, keyword=noteKeyword).one()
        except NoResultFound:
            # Already a task
            return
        session.delete(taskKeyword)

    def isNote(self, session):
        noteKeyword = Task.getNoteKeyword(session)
        return any((x.keyword == noteKeyword for x in self.taskKeywords))

    def __repr__(self):
        return "<Task id={} title={}>".format(self.id, self.title)


class Config(Base):
    """yokadi config"""
    __tablename__ = "config"
    id = Column(Integer, primary_key=True)
    name = Column(Unicode, unique=True)
    value = Column(Unicode)
    system = Column(Boolean)
    desc = Column(Unicode)


class TaskLock(Base):
    __tablename__ = "task_lock"
    id = Column(Integer, primary_key=True)
    taskId = Column("task_id", Integer, ForeignKey("task.id"), unique=True, nullable=False)
    pid = Column(Integer, default=None)
    updateDate = Column("update_date", DateTime, default=None)


class Alias(Base):
    __tablename__ = "alias"
    uuid = Column(Unicode, unique=True, default=uuidGenerator, nullable=False, primary_key=True)
    name = Column(Unicode, unique=True, nullable=False)
    command = Column(Unicode, nullable=False)

    @staticmethod
    def getAsDict(session):
        dct = {}
        for alias in session.query(Alias).all():
            dct[alias.name] = alias.command
        return dct

    @staticmethod
    def add(session, name, command):
        alias = Alias(name=name, command=command)
        session.add(alias)

    @staticmethod
    def rename(session, name, newName):
        alias = session.query(Alias).filter_by(name=name).one()
        alias.name = newName

    @staticmethod
    def setCommand(session, name, command):
        alias = session.query(Alias).filter_by(name=name).one()
        alias.command = command


def getConfigKey(name, environ=True):
    session = getSession()
    if environ:
        return os.environ.get(name, session.query(Config).filter_by(name=name).one().value)
    else:
        return session.query(Config).filter_by(name=name).one().value


_database = None


def getSession():
    global _database
    if not _database:
        raise YokadiException("Cannot get session. Not connected to database")
    return _database.session


def connectDatabase(dbFileName, createIfNeeded=True, memoryDatabase=False):
    global _database
    _database = Database(dbFileName, createIfNeeded, memoryDatabase)


class Database(object):
    def __init__(self, dbFileName, createIfNeeded=True, memoryDatabase=False, updateMode=False):
        """Connect to database and create it if needed
        @param dbFileName: path to database file
        @type dbFileName: str
        @param createIfNeeded: Indicate if database must be created if it does not exists (default True)
        @type createIfNeeded: bool
        @param memoryDatabase: create db in memory. Only usefull for unit test. Default is false.
        @type memoryDatabase: bool
        @param updateMode: allow to use it without checking version. Default is false.
        @type updateMode: bool
        """

        dbFileName = os.path.abspath(dbFileName)

        if memoryDatabase:
            connectionString = "sqlite:///:memory:"
        else:
            connectionString = 'sqlite:///' + dbFileName

        echo = os.environ.get("YOKADI_SQL_DEBUG", "0") != "0"
        self.engine = create_engine(connectionString, echo=echo)
        self.session = scoped_session(sessionmaker(bind=self.engine))

        if not os.path.exists(dbFileName) or memoryDatabase:
            if not createIfNeeded:
                raise DbUserException("Database file (%s) does not exist or is not readable." % dbFileName)
            if not memoryDatabase:
                print("Creating %s" % dbFileName)
            self.createTables()
            # Set database version according to current yokadi release
            # Don't do it in updateMode: the update script adds the version from the dump
            if not updateMode:
                self.session.add(Config(name=DB_VERSION_KEY, value=str(DB_VERSION), system=True,
                                        desc="Database schema release number"))
            self.session.commit()

        if not updateMode:
            self.checkVersion()

    def createTables(self):
        """Create all defined tables"""
        Base.metadata.create_all(self.engine)

    def getVersion(self):
        if not self._hasConfigTable():
            # There was no Config table in v1
            return 1

        try:
            return int(self.session.query(Config).filter_by(name=DB_VERSION_KEY).one().value)
        except NoResultFound:
            raise YokadiException("Configuration key '%s' does not exist. This should not happen!" % DB_VERSION_KEY)

    def setVersion(self, version):
        assert self._hasConfigTable()
        instance = self.session.query(Config).filter_by(name=DB_VERSION_KEY).one()
        instance.value = str(version)
        self.session.add(instance)
        self.session.commit()

    def _hasConfigTable(self):
        inspector = inspect(self.engine)
        return inspector.has_table("config")

    def checkVersion(self):
        """Check version and exit if it is not suitable"""
        version = self.getVersion()
        if version != DB_VERSION:
            msg = "Your database version is {} but Yokadi wants version {}.\n".format(version, DB_VERSION)
            msg += "Please run Yokadi with the --update option to update your database."
            raise DbUserException(msg)


def setDefaultConfig():
    """Set default config parameter in database if they (still) do not exist"""
    defaultConfig = {
        "ALARM_DELAY_CMD":
            ('''kdialog --passivepopup "task {TITLE} ({ID}) is due for {DATE}" 180 --title "Yokadi: {PROJECT}"''',
             False, "Command executed by Yokadi Daemon when a tasks due date is reached soon (see ALARM_DELAY"),
        "ALARM_DUE_CMD":
            ('''kdialog --passivepopup "task {TITLE} ({ID}) should be done now" 1800 --title "Yokadi: {PROJECT}"''',
             False, "Command executed by Yokadi Daemon when a tasks due date is reached soon (see ALARM_DELAY"),
        "ALARM_DELAY": ("8", False, "Delay (in hours) before due date to launch the alarm (see ALARM_CMD)"),
        "ALARM_SUSPEND": ("1", False, "Delay (in hours) before an alarm trigger again"),
        "PURGE_DELAY": ("90", False, "Default delay (in days) for the t_purge command"),
    }

    session = getSession()
    for name, value in defaultConfig.items():
        if session.query(Config).filter_by(name=name).count() == 0:
            session.add(Config(name=name, value=value[0], system=value[1], desc=value[2]))
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Update from version 3 to version 4 of Yokadi DB

@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or newer
"""


def createRecurrenceTable(cursor):
    cursor.execute("""
create table recurrence (
    id integer not null,
    rule varchar,
    primary key (id)
)
""")


def addTaskRecurrenceIdColumn(cursor):
    cursor.execute("alter table task add column recurrence_id integer references recurrence(id)")


def removeDefaultProject(cursor):
    cursor.execute("delete from config where name='DEFAULT_PROJECT'")


def update(cursor):
    createRecurrenceTable(cursor)
    addTaskRecurrenceIdColumn(cursor)
    removeDefaultProject(cursor)


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Task related commands.

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
import os
import readline
import re
from datetime import datetime, timedelta
from sqlalchemy import or_, desc
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound

from yokadi.core.db import Keyword, Project, Task, TaskKeyword, NOTE_KEYWORD
from yokadi.core import bugutils
from yokadi.core import dbutils
from yokadi.core import db
from yokadi.core import ydateutils
from yokadi.core.recurrencerule import RecurrenceRule
from yokadi.ycli import massedit
from yokadi.ycli.basicparseutils import parseOneWordName
from yokadi.ycli import parseutils
from yokadi.ycli import tui
from yokadi.ycli.completers import ProjectCompleter, projectAndKeywordCompleter, \
    taskIdCompleter, recurrenceCompleter, dueDateCompleter
from yokadi.core.dbutils import DbFilter, KeywordFilter
from yokadi.core.yokadiexception import YokadiException, BadUsageException
from yokadi.ycli.textlistrenderer import TextListRenderer
from yokadi.ycli.xmllistrenderer import XmlListRenderer
from yokadi.ycli.csvlistrenderer import CsvListRenderer
from yokadi.ycli.htmllistrenderer import HtmlListRenderer
from yokadi.ycli.plainlistrenderer import PlainListRenderer
from yokadi.core.yokadioptionparser import YokadiOptionParser

gRendererClassDict = dict(
    text=TextListRenderer,
    xml=XmlListRenderer,
    csv=CsvListRenderer,
    html=HtmlListRenderer,
    plain=PlainListRenderer,
)


class TaskCmd(object):
    def __init__(self):
        self.lastTaskId = None  # Last id created, used
        self.lastProjectName = None  # Last project name used
        self.lastTaskIds = []  # Last list of ids selected with t_list
        self.kFilters = []  # Permanent keyword filters (List of KeywordFilter)
        self.pFilter = ""  # Permanent project filter (name of project)
        self.session = db.getSession()
        for name in bugutils.PROPERTY_NAMES:
            dbutils.getOrCreateKeyword(name, interactive=False)
        dbutils.getOrCreateKeyword(NOTE_KEYWORD, interactive=False)
        self.session.commit()

    def _parser_t_add(self, cmd):
        """Code shared by t_add, bug_add and n_add parsers."""
        parser = YokadiOptionParser()
        parser.usage = "%s [options] <projectName> [@<keyword1>] [@<keyword2>] <title>" % cmd
        parser.description = "Add new %s. Will prompt to create keywords if they do not exist." % cmd
        parser.add_argument("-d", "--describe", dest="describe", default=False, action="store_true",
                            help="Directly open editor to describe task")
        parser.add_argument('cmd', nargs='*')
        return parser

    def _t_add(self, cmd, line):
        """Code shared by t_add, bug_add and n_add."""
        parser = self._parser_t_add(cmd)
        args = parser.parse_args(line)

        line = " ".join(args.cmd)
        if not line:
            raise BadUsageException("Missing parameters")
        projectName, title, keywordDict = parseutils.parseLine(line)
        projectName = self._realProjectName(projectName)
        if not title:
            raise BadUsageException("Missing title")

        task = dbutils.addTask(projectName, title, keywordDict)
        if not task:
            tui.reinjectInRawInput("%s %s" % (cmd, line))
            return None
        self.lastTaskId = task.id

        if args.describe:
            self.do_t_describe(self.lastTaskId)
        return task

    def do_t_add(self, line):
        """Add new task. Will prompt to create keywords if they do not exist.
        t_add <projectName> [@<keyword1>] [@<keyword2>] <title>"""
        task = self._t_add("t_add", line)
        if task:
            self.session.add(task)
            self.session.commit()
            print("Added task '%s' (id=%d)" % (task.title, task.id))

    complete_t_add = projectAndKeywordCompleter

    def do_bug_add(self, line):
        """Add a bug-type task. Will create a task and ask additional info.
        bug_add <project_name> [@<keyword1>] [@<keyword2>] <title>
        """
        task = self._t_add("bug_add", line)
        if not task:
            return

        keywordDict = task.getKeywordDict()
        bugutils.editBugKeywords(keywordDict)
        task.setKeywordDict(keywordDict)
        task.urgency = bugutils.computeUrgency(keywordDict)

        self.session.add(task)
        self.session.commit()
        print("Added bug '%s' (id=%d, urgency=%d)" % (task.title, task.id, task.urgency))

    complete_bug_add = ProjectCompleter(1)

    def do_n_add(self, line):
        """Add a note. A note is a task with the @_note keyword.
        n_add <project_name> [@<keyword1>] [@<keyword2>] <title>
        """
        task = self._t_add("n_add", line)
        if not task:
            return
        self.session.add(task)
        keywordDict = task.getKeywordDict()
        keywordDict[NOTE_KEYWORD] = None
        task.setKeywordDict(keywordDict)
        self.session.commit()
        print("Added note '%s' (id=%d)" % (task.title, task.id))
    complete_n_add = projectAndKeywordCompleter

    def do_bug_edit(self, line):
        """Edit a bug.
        bug_edit <id>"""
        task = self._t_edit(line, keywordEditor=bugutils.editBugKeywords)
        if task:
            self.session.commit()
    complete_bug_edit = taskIdCompleter

    def getTaskFromId(self, tid):
        if tid == '_':
            if self.lastTaskId is None:
                raise YokadiException("No previous task defined")
            tid = self.lastTaskId
        task = dbutils.getTaskFromId(tid)
        if tid != '_':
            self.lastTaskId = task.id
        return task

    def do_t_describe(self, line):
        """Starts an editor to enter a longer description of a task.
        t_describe <id>"""
        def updateDescription(description):
            task.description = description

        task = self.getTaskFromId(line)
        try:
            description = tui.editText(task.description,
                                       onChanged=updateDescription,
                                       lockManager=dbutils.TaskLockManager(task),
                                       prefix="yokadi-%s-%s-" % (task.project, task.title))
        except Exception as e:
            raise YokadiException(e)
        updateDescription(description)
        self.session.commit()

    complete_t_describe = taskIdCompleter

    def do_t_set_urgency(self, line):
        """@deprecated: should be removed"""
        tui.warnDeprecated("t_set_urgency", "t_urgency")
        self.do_t_urgency(line)

    def do_t_urgency(self, line):
        """Defines urgency of a task.
        t_urgency <id> <value>"""

        tokens = parseutils.simplifySpaces(line).split(" ")
        if len(tokens) != 2:
            raise BadUsageException("You must provide a taskId and an urgency value")
        task = self.getTaskFromId(tokens[0])
        try:
            # Do not use isdigit(), so that we can set negative urgency. This
            # make it possible to stick tasks to the bottom of the list.
            urgency = int(tokens[1])
        except ValueError:
            raise BadUsageException("Task urgency must be a digit")

        if urgency > 100:
            tui.warning("Max urgency is 100")
            urgency = 100
        elif urgency < -99:
            tui.warning("Min urgency is -99")
            urgency = -99

        task.urgency = urgency
        self.session.commit()

    complete_t_set_urgency = taskIdCompleter
    complete_t_urgency = taskIdCompleter

    def do_t_mark_started(self, line):
        """Mark task as started.
        t_mark_started <id>"""
        self._t_set_status(line, 'started')

    complete_t_mark_started = taskIdCompleter

    def do_t_mark_done(self, line):
        """Mark task as done.
        t_mark_done <id>"""
        self._t_set_status(line, 'done')

    complete_t_mark_done = taskIdCompleter

    def do_t_mark_new(self, line):
        """Mark task as new (not started).
        t_mark_new <id>"""
        self._t_set_status(line, 'new')

    complete_t_mark_new = taskIdCompleter

    def _t_set_status(self, line, status):
        task = self.getTaskFromId(line)
        task.setStatus(status)
        self.session.commit()
        if task.recurrence and status == "done":
            print("Task '%s' next occurrence is scheduled at %s" % (task.title, task.dueDate))
            print("To *really* mark this task done and forget it, remove its recurrence first"
                  " with t_recurs %s none" % task.id)
        else:
            print("Task '%s' marked as %s" % (task.title, status))

    def do_t_apply(self, line):
        """Apply a command to several tasks.
        t_apply <id1>[,<id2>,[<id3>]...]] <command> <args>
        Use x-y to select task range from x to y
        Use __ to select all tasks previously selected with t_list"""
        ids = []
        if "__" in line:
            if self.lastTaskIds:
                line = line.replace("__", ",".join([str(i) for i in self.lastTaskIds]))
            else:
                raise BadUsageException("You must select tasks with t_list prior to use __")
        rangeId = re.compile(r"(\d+)-(\d+)")
        tokens = re.split(r"[\s|,]", line)
        if len(tokens) < 2:
            raise BadUsageException("Give at least a task id and a command")

        idScan = True  # Indicate we are parsing ids
        cmdTokens = []  # Command that we want to apply
        for token in tokens:
            if token == "":
                continue
            if idScan:
                result = rangeId.match(token)
                if result:
                    ids.extend(list(range(int(result.group(1)), int(result.group(2)) + 1)))
                elif token.isdigit():
                    ids.append(int(token))
                else:
                    # Id list is finished. Grab rest of line.
                    cmdTokens.append(token)
                    idScan = False
            else:
                cmdTokens.append(token)

        if not cmdTokens:
            raise BadUsageException("Give a command to apply")
        cmd = cmdTokens.pop(0)
        for taskId in ids:
            line = " ".join([cmd, str(taskId), " ".join(cmdTokens)])
            print("Executing: %s" % line)
            self.onecmd(line.strip())

    complete_t_apply = taskIdCompleter

    def parser_t_remove(self):
        parser = YokadiOptionParser()
        parser.usage = "t_remove [options] <id>"
        parser.description = "Delete a task."
        parser.add_argument("-f", dest="force", default=False, action="store_true",
                            help="Skip confirmation prompt")
        parser.add_argument("id")
        return parser

    def do_t_remove(self, line):
        parser = self.parser_t_remove()
        args = parser.parse_args(line)
        task = self.getTaskFromId(args.id)
        if not args.force:
            if not tui.confirm("Remove task '%s'" % task.title):
                return
        project = task.project
        self.session.delete(task)
        print("Task '%s' removed" % (task.title))

        # Delete project with no associated tasks
        if self.session.query(Task).filter_by(project=project).count() == 0:
            self.session.delete(project)
        self.session.commit()
    complete_t_remove = taskIdCompleter

    def parser_t_purge(self):
        parser = YokadiOptionParser()
        parser.usage = "t_purge [options]"
        parser.description = "Remove old done tasks from all projects."
        parser.add_argument("-f", "--force", dest="force", default=False, action="store_true",
                            help="Skip confirmation prompt")
        delay = int(db.getConfigKey("PURGE_DELAY", environ=False))
        parser.add_argument("-d", "--delay", dest="delay", default=delay,
                            type=int, help="Delay (in days) after which done tasks are destroyed."
                                           " Default is %d." % delay)
        return parser

    def do_t_purge(self, line):
        parser = self.parser_t_purge()
        args = parser.parse_args(line)
        filters = []
        filters.append(Task.status == "done")
        filters.append(Task.doneDate < (datetime.now() - timedelta(days=args.delay)))
        tasks = self.session.query(Task).filter(*filters)
        if tasks.count() == 0:
            print("No tasks need to be purged")
            return
        print("The following tasks will be removed:")
        print("\n".join(["%s: %s" % (task.id, task.title) for task in tasks]))
        if args.force or tui.confirm("Do you really want to remove those tasks (this action cannot be undone)?"):
            self.session.delete(tasks)
            self.session.commit()
            print("Tasks deleted")
        else:
            print("Purge canceled")

    def parser_t_list(self):
        parser = YokadiOptionParser()
        parser.usage = "t_list [options] <project_or_keyword_filter>"
        parser.description = "List tasks filtered by project and/or keywords. " \
                             "'%' can be used as a wildcard in the project name: " \
                             "to list projects starting with 'foo', use 'foo%'. " \
                             "Keyword filtering is achieved with '@'. Ex.: " \
                             "t_list @home, t_list @_bug=2394"

        parser.add_argument("-a", "--all", dest="status",
                            action="store_const", const="all",
                            help="all tasks (done and to be done)")

        parser.add_argument("--started", dest="status",
                            action="store_const", const="started",
                            help="only started tasks")

        rangeList = ["today", "thisweek", "thismonth", "all"]
        parser.add_argument("-d", "--done", dest="done",
                            help="only done tasks. <range> must be either one of %s or a date using the same format"
                                 " as t_due" % ", ".join(rangeList),
                            metavar="<range>")

        parser.add_argument("-u", "--urgency", dest="urgency",
                            type=int,
                            help="tasks with urgency greater or equal than <urgency>",
                            metavar="<urgency>")

        parser.add_argument("-t", "--top-due", dest="topDue",
                            default=False, action="store_true",
                            help="top 5 urgent tasks of each project based on due date")

        parser.add_argument("--overdue", dest="due",
                            action="append_const", const="now",
                            help="all overdue tasks")

        parser.add_argument("--due", dest="due",
                            action="append",
                            help="""only list tasks due before/after <limit>. <limit> is a
                            date optionaly prefixed with a comparison operator.
                            Valid operators are: <, <=, >=, and >.
                            Example of valid limits:

                            - tomorrow: due date <= tomorrow, 23:59:59
                            - today: due date <= today, 23:59:59
                            - >today: due date > today: 23:59:59
                            """,
                            metavar="<limit>")

        parser.add_argument("-k", "--keyword", dest="keyword",
                            help="Group tasks by given keyword instead of project. The %% wildcard can be used.",
                            metavar="<keyword>")

        parser.add_argument("-s", "--search", dest="search",
                            action="append",
                            help="only list tasks whose title or description match <value>. You can repeat this"
                                 " option to search on multiple words.",
                            metavar="<value>")

        formatList = ["auto"] + list(gRendererClassDict.keys())
        parser.add_argument("-f", "--format", dest="format",
                            default="auto", choices=formatList,
                            help="how should the task list be formated. <format> can be %s" % ", ".join(formatList),
                            metavar="<format>")
        parser.add_argument("-o", "--output", dest="output",
                            help="Output task list to <file>",
                            metavar="<file>")

        parser.add_argument("filter", nargs="*", metavar="<project_or_keyword_filter>")

        return parser

    def _realProjectName(self, name):
        if name == '_':
            if self.lastProjectName is None:
                raise YokadiException("No previous project used")
        else:
            self.lastProjectName = name
        return self.lastProjectName

    def _parseListLine(self, parser, line):
        """
        Parse line with parser, returns a tuple of the form
        (options, projectList, filters)
        """
        args = parser.parse_args(line)
        if len(args.filter) > 0:
            projectName, filters = parseutils.extractKeywords(" ".join(args.filter))
        else:
            projectName = ""
            filters = []

        if self.kFilters:
            # Add keyword filter
            filters.extend(self.kFilters)

        if not projectName:
            if self.pFilter:
                # If a project filter is defined, use it as none was provided
                projectName = self.pFilter
            else:
                # Take all project if none provided
                projectName = "%"

        if projectName.startswith("!"):
            projectName = self._realProjectName(projectName[1:])
            projectList = self.session.query(Project).filter(Project.name.notlike(projectName)).all()
        else:
            projectName = self._realProjectName(projectName)
            projectList = self.session.query(Project).filter(Project.name.like(projectName)).all()

        if len(projectList) == 0:
            raise YokadiException("Found no project matching '%s'" % projectName)

        # Check keywords exist
        parseutils.warnIfKeywordDoesNotExist(filters)

        # Search
        if args.search:
            for word in args.search:
                if word.startswith("@"):
                    tui.warning("Maybe you want keyword search (without -s option) "
                                "instead of plain text search?")
                condition = or_(Task.title.like("%" + word + "%"),
                                Task.description.like("%" + word + "%"))
                filters.append(DbFilter(condition))

        return args, projectList, filters

    def _renderList(self, renderer, projectList, filters, order,
                    limit=None, groupKeyword=None):
        """
        Render a list using renderer, according to the restrictions set by the
        other parameters
        @param renderer: renderer class (for example: TextListRenderer)
        @param projectList: list of project name (as unicode string)
        @param filters: filters in sql alchemy format (example: Task.status == 'done')
        @param order: ordering in sqlalchemy format (example: desc(Task.urgency))
        @param limit: limit number tasks (int) or None for no limit
        @param groupKeyword: keyword used for grouping (as unicode string) or None
        """
        def applyFilters(lst):
            for filter in filters:
                lst = filter.apply(lst)
            return lst

        if groupKeyword:
            if groupKeyword.startswith("@"):
                groupKeyword = groupKeyword[1:]
            keywords = self.session.query(Keyword).filter(Keyword.name.like(groupKeyword))

            for keyword in sorted(keywords, key=lambda x: x.name.lower()):
                if str(keyword.name).startswith("_") and not groupKeyword.startswith("_"):
                    # BUG: cannot filter on db side because sqlobject does not
                    # understand ESCAPE needed with _. Need to test it with
                    # sqlalchemy
                    continue
                taskList = self.session.query(Task).filter(TaskKeyword.keywordId == keyword.id)
                taskList = taskList.outerjoin(TaskKeyword, Task.taskKeywords)
                taskList = applyFilters(taskList)
                taskList = taskList.order_by(*order).limit(limit).distinct()
                taskList = list(taskList)
                if projectList:
                    taskList = [x for x in taskList if x.project in projectList]
                if len(taskList) > 0:
                    self.lastTaskIds.extend([t.id for t in taskList])  # Keep selected id for further use
                    renderer.addTaskList(str(keyword), taskList)
            renderer.end()
        else:
            hiddenProjectNames = []
            for project in sorted(projectList, key=lambda x: x.name.lower()):
                if not project.active:
                    hiddenProjectNames.append(project.name)
                    continue
                taskList = self.session.query(Task).filter(Task.project == project)
                taskList = taskList.outerjoin(TaskKeyword, Task.taskKeywords)
                taskList = applyFilters(taskList)
                taskList = taskList.order_by(*order).limit(limit).distinct()
                taskList = list(taskList)
                if len(taskList) > 0:
                    self.lastTaskIds.extend([t.id for t in taskList])  # Keep selected id for further use
                    renderer.addTaskList(str(project), taskList)
            renderer.end()

            if len(hiddenProjectNames) > 0:
                tui.info("hidden projects: %s" % ", ".join(hiddenProjectNames))

    def do_t_list(self, line, renderer=None):

        def selectRendererClass():
            if args.format != "auto":
                return gRendererClassDict[args.format]

            defaultRendererClass = TextListRenderer
            if not args.output:
                return defaultRendererClass

            ext = os.path.splitext(args.output)[1]
            if not ext:
                return defaultRendererClass

            return gRendererClassDict.get(ext[1:], defaultRendererClass)

        # Reset last tasks id list
        self.lastTaskIds = []

        # BUG: completion based on parameter position is broken when parameter is given"
        args, projectList, filters = self._parseListLine(self.parser_t_list(), line)

        # Skip notes
        filters.append(KeywordFilter(NOTE_KEYWORD, negative=True))

        # Handle t_list specific options
        order = [desc(Task.urgency), Task.creationDate]
        limit = None
        if args.done:
            filters.append(DbFilter(Task.status == 'done'))
            if args.done != "all":
                minDate = ydateutils.parseMinDate(args.done)
                filters.append(DbFilter(Task.doneDate >= minDate))
        elif args.status == "all":
            pass
        elif args.status == "started":
            filters.append(DbFilter(Task.status == "started"))
        else:
            filters.append(DbFilter(Task.status != "done"))
        if args.urgency is not None:
            order = [desc(Task.urgency), ]
            filters.append(DbFilter(Task.urgency >= args.urgency))
        if args.topDue:
            filters.append(DbFilter(Task.dueDate is not None))
            order = [Task.dueDate, ]
            limit = 5
        if args.due:
            for due in args.due:
                dueOperator, dueLimit = ydateutils.parseDateLimit(due)
                filters.append(DbFilter(dueOperator(Task.dueDate, dueLimit)))
            order = [Task.dueDate, ]

        # Define output
        if args.output:
            out = open(args.output, "w", encoding='utf-8')
        else:
            out = tui.stdout

        # Instantiate renderer
        if renderer is None:
            rendererClass = selectRendererClass()
            renderer = rendererClass(out)

        # Fill the renderer
        self._renderList(renderer, projectList, filters, order, limit, args.keyword)
    complete_t_list = projectAndKeywordCompleter

    def parser_n_list(self):
        parser = YokadiOptionParser()
        parser.usage = "n_list [options] <project_or_keyword_filter>"
        parser.description = "List notes filtered by project and/or keywords. " \
            "'%' can be used as a wildcard in the project name: " \
            "to list projects starting with 'foo', use 'foo%'. " \
            "Keyword filtering is achieved with '@'. Ex.: " \
            "n_list @home, n_list @_bug=2394"

        parser.add_argument("-s", "--search", dest="search",
                            action="append",
                            help="only list notes whose title or description match <value>."
                                 " You can repeat this option to search on multiple words.",
                            metavar="<value>")

        parser.add_argument("-k", "--keyword", dest="keyword",
                            help="Group tasks by given keyword instead of project. The '%%' wildcard can be used.",
                            metavar="<keyword>")
        parser.add_argument("filter", nargs="*", metavar="<project_or_keyword_filter>")

        return parser

    def do_n_list(self, line):
        args, projectList, filters = self._parseListLine(self.parser_n_list(), line)

        filters.append(KeywordFilter(NOTE_KEYWORD))
        order = [Task.creationDate, ]
        renderer = TextListRenderer(tui.stdout, renderAsNotes=True)
        self._renderList(renderer, projectList, filters, order, limit=None,
                         groupKeyword=args.keyword)
    complete_n_list = projectAndKeywordCompleter

    def do_t_reorder(self, line):
        """Reorder tasks of a project.
        It works by starting an editor with the task list: you can then change
        the order of the lines and save the list. The urgency field will be
        updated to match the order.
        t_reorder <project_name>"""
        try:
            project = self.session.query(Project).filter_by(name=line).one()
        except (MultipleResultsFound, NoResultFound):
            raise BadUsageException("You must provide a valid project name")

        taskList = self.session.query(Task).filter(Task.projectId == project.id,
                                                   Task.status != 'done').order_by(desc(Task.urgency))
        lines = ["%d,%s" % (x.id, x.title) for x in taskList]
        text = tui.editText("\n".join(lines))

        ids = []
        for line in text.split("\n"):
            line = line.strip()
            if "," not in line:
                continue
            ids.append(int(line.split(",")[0]))

        ids.reverse()
        for urgency, taskId in enumerate(ids):
            task = self.session.get(Task, taskId)
            task.urgency = urgency
        self.session.commit()
    complete_t_reorder = ProjectCompleter(1)

    def do_t_medit(self, line):
        """Mass edit tasks of a project.
        t_medit <project_name>

        Starts a text editor with the task list, you can then:
        - edit tasks text and keywords
        - mark tasks as done or started
        - add new tasks
        - adjust urgency
        - delete tasks
        """
        if not line:
            raise BadUsageException("Missing parameters")
        projectName = parseOneWordName(line)
        projectName = self._realProjectName(projectName)
        project = dbutils.getOrCreateProject(projectName)
        if not project:
            return

        oldList = massedit.createEntriesForProject(project)
        oldText = massedit.createMEditText(oldList)
        newText = oldText
        while True:
            newText = tui.editText(newText, suffix=".medit")
            if newText == oldText:
                print("No changes")
                return

            try:
                newList = massedit.parseMEditText(newText)
            except massedit.ParseError as exc:
                print(exc)
                print()
                if tui.confirm("Modify text and try again"):
                    lst = newText.splitlines()
                    lst.insert(exc.lineNumber, "# ^ " + exc.message)
                    newText = "\n".join(lst)
                    continue
                else:
                    return

            try:
                massedit.applyChanges(project, oldList, newList)
                self.session.commit()
                break
            except YokadiException as exc:
                print(exc)
                print()
                if not tui.confirm("Modify text and try again"):
                    return
    complete_t_medit = ProjectCompleter(1)

    def parser_t_show(self):
        parser = YokadiOptionParser()
        parser.usage = "t_show [options] <id>"
        parser.description = "Display details of a task."
        choices = ["all", "summary", "description"]
        parser.add_argument("--output", dest="output",
                            choices=choices,
                            default="all",
                            help="<output> can be one of %s. If not set, it defaults to all." % ", ".join(choices),
                            metavar="<output>")
        parser.add_argument("id")
        return parser

    def do_t_show(self, line):
        parser = self.parser_t_show()
        args = parser.parse_args(line)

        task = self.getTaskFromId(args.id)

        if args.output in ("all", "summary"):
            keywordDict = task.getKeywordDict()
            keywordArray = []
            for name, value in list(keywordDict.items()):
                txt = name
                if value:
                    txt += "=" + str(value)
                keywordArray.append(txt)
                keywordArray.sort()
            keywords = ", ".join(keywordArray)

            if task.recurrence:
                recurrence = "{} (next: {})".format(
                    task.recurrence.getFrequencyAsString(),
                    task.recurrence.getNext()
                )
            else:
                recurrence = "None"

            fields = [
                ("Project", task.project.name),
                ("Title", task.title),
                ("ID", task.id),
                ("Created", task.creationDate),
                ("Due", task.dueDate),
                ("Status", task.status),
                ("Urgency", task.urgency),
                ("Recurrence", recurrence),
                ("Keywords", keywords),
            ]

            if task.status == "done":
                fields.append(("Done", task.doneDate))

            tui.renderFields(fields)

        if args.output in ("all", "description") and task.description:
            if args.output == "all":
                print()
            print(task.description)

    complete_t_show = taskIdCompleter

    def _t_edit(self, line, keywordEditor=None):
        """Code shared by t_edit and bug_edit.
        if keywordEditor is not None it will be called after editing the task.
        Returns the modified task if OK, None if cancelled"""
        def editComplete(text, state):
            """ Specific completer for the edit prompt.
            This subfunction should stay here because it needs to access to cmd members"""
            if state == 0:
                origline = readline.get_line_buffer()
                line = origline.lstrip()
                stripped = len(origline) - len(line)
                begidx = readline.get_begidx() - stripped
                endidx = readline.get_endidx() - stripped
                if begidx > 0:
                    self.completion_matches = projectAndKeywordCompleter("", text, line, begidx, endidx, shift=1)
                else:
                    self.completion_matches = []
            try:
                return self.completion_matches[state]
            except IndexError:
                return None

        task = self.getTaskFromId(line)

        # Create task line
        keywordDict = task.getKeywordDict()
        userKeywordDict, keywordDict = dbutils.splitKeywordDict(keywordDict)
        taskLine = parseutils.createLine("", task.title, userKeywordDict)

        oldCompleter = readline.get_completer()  # Backup previous completer to restore it in the end
        readline.set_completer(editComplete)  # Switch to specific completer

        # Edit
        try:
            while True:
                print("(Press Ctrl+C to cancel)")
                try:
                    line = tui.editLine(taskLine)
                    if not line.strip():
                        tui.warning("Missing title")
                        continue
                except KeyboardInterrupt:
                    print()
                    print("Cancelled")
                    return None
                _, title, userKeywordDict = parseutils.parseLine(task.project.name + " " + line)

                if dbutils.createMissingKeywords(userKeywordDict.keys()):
                    # We were able to create missing keywords if there were any,
                    # we can now exit the edit loop
                    break
        finally:
            readline.set_completer(oldCompleter)

        keywordDict.update(userKeywordDict)
        if keywordEditor:
            keywordEditor(keywordDict)

        task.title = title
        task.setKeywordDict(keywordDict)
        return task

    def do_t_edit(self, line):
        """Edit a task.
        t_edit <id>"""
        self._t_edit(line)
        self.session.commit()
    complete_t_edit = taskIdCompleter

    def do_t_set_project(self, line):
        """@deprecated: should be removed"""
        tui.warnDeprecated("t_set_project", "t_project")
        self.do_t_project(line)

    def do_t_project(self, line):
        """Set task's project.
        t_project <id> <project>"""
        tokens = parseutils.simplifySpaces(line).split(" ")
        if len(tokens) != 2:
            raise YokadiException("You should give two arguments: <task id> <project>")
        projectName = tokens[1]
        projectName = self._realProjectName(projectName)
        project = dbutils.getOrCreateProject(projectName)
        if not project:
            return

        task = self.getTaskFromId(tokens[0])
        task.project = project
        self.session.commit()
        if task.project:
            print("Moved task '%s' to project '%s'" % (task.title, projectName))

    complete_t_set_project = ProjectCompleter(2)
    complete_t_project = ProjectCompleter(2)

    def do_t_set_due(self, line):
        """@deprecated: should be removed"""
        tui.warnDeprecated("t_set_due", "t_due")
        self.do_t_due(line)

    def do_t_due(self, line):
        """Set task's due date
        t_due <id> <date>

        Date can be specified as a relative offset:
        - +5M: in 5 minutes
        - +3H: in 3 hours
        - +1D: in 1 day
        - +6W: in 6 weeks

        As a day in the week:
        - tomorrow:      tomorrow, same hour
        - tuesday 12:10: next tuesday, at 12:10
        - fr 15:30:      next friday, at 15:30

        Or as an absolute date or time:
        - 10:38:            at 10:38 today
        - 25/09/2010 12:10: on the 25th of September, 2010, at 12:10
        - 23/02/2010:       on the 23th of February, 2010
        - 01/04:            on the 1st of April
        - 12:               on the 12th of current month

        To reset a due date, use "none"."""
        line = parseutils.simplifySpaces(line)
        if len(line.split()) < 2:
            raise YokadiException("Give a task id and time, date or date & time")
        taskId, line = line.strip().split(" ", 1)
        task = self.getTaskFromId(taskId)

        if line.lower() == "none":
            task.dueDate = None
            print("Due date for task '%s' reset" % task.title)
        else:
            task.dueDate = ydateutils.parseHumaneDateTime(line)
            print("Due date for task '%s' set to %s" % (task.title, task.dueDate.ctime()))
        self.session.commit()
    complete_t_set_due = dueDateCompleter
    complete_t_due = dueDateCompleter

    def do_t_add_keywords(self, line):
        """Add keywords to an existing task
        t_add_keywords <id> <@keyword1> <@keyword2>[=<value>]...
        """
        tokens = parseutils.simplifySpaces(line).split(" ", 1)
        if len(tokens) < 2:
            raise YokadiException("You should give at least two arguments: <task id> <keyword>")
        task = dbutils.getTaskFromId(tokens[0])
        garbage, keywordFilters = parseutils.extractKeywords(tokens[1])
        newKwDict = parseutils.keywordFiltersToDict(keywordFilters)
        if garbage:
            raise YokadiException("Cannot parse line, got garbage (%s). Maybe you forgot to add @ before keyword ?"
                                  % garbage)

        if not dbutils.createMissingKeywords(list(newKwDict.keys())):
            # User cancel keyword creation
            return

        kwDict = task.getKeywordDict()
        kwDict.update(newKwDict)
        task.setKeywordDict(kwDict)
        self.session.commit()

    def do_t_recurs(self, line):
        """Make a task recurs
        t_recurs <id> yearly <dd/mm> <HH:MM>
        t_recurs <id> monthly <dd> <HH:MM>
        t_recurs <id> monthly <first/second/third/last> <mo, tu, we, th, fr, sa, su> <hh:mm>
        t_recurs <id> quarterly <dd> <HH:MM>
        t_recurs <id> quarterly <first/second/third/last> <mo, tu, we, th, fr, sa, su> <hh:mm>
        t_recurs <id> weekly <mo, tu, we, th, fr, sa, su> <hh:mm>
        t_recurs <id> daily <HH:MM>
        t_recurs <id> none (remove recurrence)"""
        tokens = parseutils.simplifySpaces(line).split(" ", 1)
        if len(tokens) < 2:
            raise YokadiException("You should give at least two arguments: <task id> <recurrence>")
        task = self.getTaskFromId(tokens[0])
        rule = RecurrenceRule.fromHumaneString(tokens[1])
        task.setRecurrenceRule(rule)
        self.session.commit()
    complete_t_recurs = recurrenceCompleter

    def do_t_filter(self, line):
        """Define permanent keyword filter used by t_list
        Ex.:
            - t_filter @work (filter all task that have the "work" keyword)
            - t_filter none (remove filter)"""
        # TODO: add completion

        if not line:
            raise YokadiException("You must give keyword as argument or 'none' to reset filter")

        if parseutils.simplifySpaces(line).lower() == "none":
            self.kFilters = []
            self.pFilter = ""
            self.prompt = "yokadi> "
        else:
            projectName, keywordFilters = parseutils.extractKeywords(line)
            self.kFilters = keywordFilters
            self.pFilter = projectName
            prompt = "y"
            if self.pFilter:
                prompt += " %s" % projectName
            if self.kFilters:
                parseutils.warnIfKeywordDoesNotExist(self.kFilters)
                prompt += " %s" % (" ".join([str(k) for k in keywordFilters]))
            self.prompt = "%s> " % prompt

    def do_t_to_note(self, line):
        """Turns a task into a note
        """
        task = self.getTaskFromId(line)
        task.toNote(self.session)
        self.session.commit()

    def do_n_to_task(self, line):
        """Turns a note into a task
        """
        task = self.getTaskFromId(line)
        task.toTask(self.session)
        self.session.commit()

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Update from version 1 to version 2 of Yokadi DB

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""


def createConfigTable(cursor):
    cursor.execute("""create table config (
        id integer not null,
        name varchar,
        value varchar,
        system boolean,
        "desc" varchar,
        primary key (id),
        unique (name),
        check (system in (0, 1))
);""")

    rows = [
        ("DB_VERSION", "2", True, "Database schema release number"),
        ("TEXT_WIDTH", "60", False, "Width of task display output with t_list command"),
        ("DEFAULT_PROJECT", "default", False, "Default project used when no project name given"),
        ("ALARM_DELAY_CMD", '''kdialog --sorry "task {TITLE} ({ID}) is due for {DATE}" --title "Yokadi Daemon"''',
         False, "Command executed by Yokadi Daemon when a tasks due date is reached soon (see ALARM_DELAY"),
        ("ALARM_DUE_CMD", '''kdialog --error "task {TITLE} ({ID}) should be done now" --title "Yokadi Daemon"''',
         False, "Command executed by Yokadi Daemon when a tasks due date is reached soon (see ALARM_DELAY")
    ]
    for name, value, system, desc in rows:
        system = 1 if system else 0
        cursor.execute("insert into config(name, value, system, \"desc\")\n"
                       "values (?, ?, ?, ?)", (name, value, system, desc))


def addProjectActiveColumn(cursor):
    cursor.execute("alter table project add column active boolean")
    cursor.execute("update project set active = 1")


def addTableDueDateColumn(cursor):
    cursor.execute("alter table task add column due_date datetime")


def update(cursor):
    createConfigTable(cursor)
    addProjectActiveColumn(cursor)
    addTableDueDateColumn(cursor)


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Configuration management related commands.

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
from yokadi.core import db

from yokadi.core.db import Config
from yokadi.core.yokadiexception import YokadiException, BadUsageException
from yokadi.ycli.completers import confCompleter
from yokadi.ycli import tui
from yokadi.core.yokadioptionparser import YokadiOptionParser


class ConfCmd(object):
    def parser_c_get(self):
        parser = YokadiOptionParser(prog="c_get")
        parser.description = "Display the value of a configuration key. If no key is given, all keys are shown."
        parser.add_argument("-s", dest="system", default=False, action="store_true",
                            help="Display value of system keys instead of user ones")
        parser.add_argument("key", nargs='?')
        return parser

    def do_c_get(self, line):
        parser = self.parser_c_get()
        args = parser.parse_args(line)
        key = args.key
        if not key:
            key = "%"
        session = db.getSession()
        k = session.query(Config).filter(Config.name.like(key)).filter_by(system=args.system).all()
        fields = [(x.name, "%s (%s)" % (x.value, x.desc)) for x in k]
        if fields:
            tui.renderFields(fields)
        else:
            raise YokadiException("Configuration key %s does not exist" % line)

    complete_c_get = confCompleter

    def do_c_set(self, line):
        """Set a configuration key to value : c_set <key> <value>"""
        line = line.split()
        if len(line) < 2:
            raise BadUsageException("You should provide two arguments : the parameter key and the value")
        name = line[0]
        value = " ".join(line[1:])
        session = db.getSession()
        p = session.query(Config).filter_by(name=name, system=False)
        if p.count() == 0:
            raise YokadiException("Sorry, no parameter match")
        else:
            if self.checkParameterValue(name, value):
                p[0].value = value
                tui.info("Parameter updated")
            else:
                raise YokadiException("Parameter value is incorrect")

    complete_c_set = confCompleter

    def checkParameterValue(self, name, value):
        """Control that the value if ok for a parameter
        @param key: parameter name
        @param value: parameter value
        @return: True if parameter is ok, else False"""
        # Positive int parameters
        if name in ("ALARM_DELAY", "ALARM_SUSPEND", "PURGE_DELAY"):
            try:
                value = int(value)
                assert value >= 0
                return True
            except (ValueError, AssertionError):
                return False
        else:
            # No check for this parameter, so tell everything is fine
            return True



# -*- coding: UTF-8 -*-
"""
Update from version 9 to version 10 of Yokadi DB

- Remove Task.recurrence_id column
- Add Task.recurrence column
- Import recurrence from the Recurrence table
- Remove Recurrence table

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""
import json
import pickle

from yokadi.update import updateutils


def tuplify(value):
    if value is None:
        return ()
    if isinstance(value, int):
        return (value,)
    else:
        return tuple(value)


def createByweekdayValue(rule):
    if rule._bynweekday:
        # Special case: recurrence every 1st, 2nd, 3rd, 4th or last $weekday of month
        weekday, pos = rule._bynweekday[0]
        return dict(weekday=weekday, pos=pos)

    return tuplify(rule._byweekday)


def createJsonStringFromRule(pickledRule):
    try:
        rule = pickle.loads(pickledRule)
    except UnicodeDecodeError:
        # Some rules fails to unpickle if we don't set encoding to latin1
        rule = pickle.loads(pickledRule, encoding="latin1")
    dct = {}
    dct["freq"] = rule._freq
    dct["bymonth"] = tuplify(rule._bymonth)
    dct["bymonthday"] = tuplify(rule._bymonthday)
    dct["byweekday"] = createByweekdayValue(rule)
    dct["byhour"] = tuplify(rule._byhour)
    dct["byminute"] = tuplify(rule._byminute)
    return json.dumps(dct)


def addRecurrenceColumn(cursor):
    cursor.execute("alter table task add column recurrence")
    sql = "select t.id, r.rule from task t left join recurrence r on t.recurrence_id = r.id"
    for row in cursor.execute(sql).fetchall():
        id, pickledRule = row
        ruleStr = ""
        if pickledRule:
            try:
                ruleStr = createJsonStringFromRule(bytes(pickledRule, "utf-8"))
            except Exception as exc:
                print("Failed to import recurrence for task {}: {}".format(id, exc))

        cursor.execute("update task set recurrence = ? where id = ?", (ruleStr, id))


def deleteRecurrenceTable(cursor):
    cursor.execute("drop table recurrence")


def update(cursor):
    addRecurrenceColumn(cursor)
    deleteRecurrenceTable(cursor)
    updateutils.deleteTableColumns(cursor, "task", ("recurrence_id",))


if __name__ == "__main__":
    updateutils.main(update)
# vi: ts=4 sw=4 et



# coding:utf-8
"""
Ical utils functions
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""

import icalendar


def _clamp(value, minimum, maximum):
    return minimum if value < minimum else maximum if value > maximum else value


def convertIcalType(attr):
    """Convert data from icalendar types (vDates, vInt etc.) to python standard equivalent
    @param attr: icalendar type
    @return: python type"""
    if isinstance(attr, (icalendar.vDate, icalendar.vDatetime,
                         icalendar.vDuration, icalendar.vDDDTypes)):
        return attr.dt
    elif isinstance(attr, (icalendar.vInt, icalendar.vFloat)):
        return int(attr)
    else:
        # Default to unicode string
        return str(attr)


def icalPriorityToYokadiUrgency(priority):
    """Convert ical priority (1 / 9) to yokadi urgency (100 / -99)
    @param priority: ical priority
    @return: yokadi urgency"""
    urgency = 100 - 20 * priority
    return _clamp(urgency, -99, 100)


def yokadiUrgencyToIcalPriority(urgency):
    """Convert yokadi urgency (100 / -99) to ical priority (1 / 9)
    @param urgency: yokadi urgency
    @return: ical priority"""
    priority = int(-(urgency - 100) / 20)
    return _clamp(priority, 1, 9)

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Yokadi iCalendar interface

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""

import sys

try:
    import icalendar
except ImportError:
    print("You don't have the icalendar package.")
    print("Get it on http://pypi.python.org/pypi/icalendar/")
    print("Or use 'easy_install icalendar'")
    sys.exit(1)

if not hasattr(icalendar.Calendar, "from_ical"):
    print("Your version of icalendar is outdated: you need icalendar > 3.0.")
    sys.exit(1)

import http.server
from threading import Thread
import re

from yokadi.core import db
from yokadi.core.db import Task, Project
from yokadi.core import dbutils
from yokadi.yical import icalutils
from yokadi.ycli import parseutils
from yokadi.core.yokadiexception import YokadiException

# UID pattern
UID_PREFIX = "yokadi"
TASK_UID = UID_PREFIX + "-task-%s"
TASK_RE = re.compile(TASK_UID.replace("%s", r"(\d+)"))
PROJECT_UID = UID_PREFIX + "-project-%s"

# Default project where new task are added
# TODO: make this a configurable items via c_set
INBOX_PROJECT = "inbox"

# Yokadi task <=> iCalendar VTODO attribute mapping
YOKADI_ICAL_ATT_MAPPING = {"title": "summary",
                           "urgency": "priority",
                           "creationDate": "dtstart",
                           "dueDate": "due",
                           "doneDate": "completed",
                           "description": "description"}


def generateCal():
    """Generate an ical calendar from yokadi database
    @return: icalendar.Calendar object"""
    session = db.getSession()
    cal = icalendar.Calendar()
    cal.add("prodid", '-//Yokadi calendar //yokadi.github.io//')
    cal.add("version", "2.0")
    # Add projects
    for project in session.query(Project).filter(Project.active == True):  # noqa
        vTodo = icalendar.Todo()
        vTodo.add("summary", project.name)
        vTodo["uid"] = PROJECT_UID % project.id
        cal.add_component(vTodo)
    # Add tasks
    for task in session.query(Task).filter(Task.status != "done"):
        vTodo = createVTodoFromTask(task)
        cal.add_component(vTodo)

    return cal


def createVTodoFromTask(task):
    """Create a VTodo object from a yokadi task
    @param task: yokadi task (db.Task object)
    @return: ical VTODO (icalendar.Calendar.Todo object)"""
    vTodo = icalendar.Todo()
    vTodo["uid"] = TASK_UID % task.id
    vTodo["related-to"] = PROJECT_UID % task.project.id

    # Add standard attribute
    for yokadiAttribute, icalAttribute in list(YOKADI_ICAL_ATT_MAPPING.items()):
        attr = getattr(task, yokadiAttribute)
        if attr:
            if yokadiAttribute == "urgency":
                attr = icalutils.yokadiUrgencyToIcalPriority(attr)
            if yokadiAttribute == "title":
                attr += " (%s)" % task.id
            vTodo.add(icalAttribute, attr)

    # Add categories from keywords
    categories = []
    if task.taskKeywords:
        for name, value in list(task.getKeywordDict().items()):
            if value:
                categories.append("%s=%s" % (name, value))
            else:
                categories.append(name)
    vTodo.add("categories", categories)

    return vTodo


def updateTaskFromVTodo(task, vTodo):
    """Update a yokadi task with an ical VTODO object
    @param task: yokadi task (db.Task object)
    @param vTodo: ical VTODO (icalendar.Calendar.Todo object)"""
    for yokadiAttribute, icalAttribute in list(YOKADI_ICAL_ATT_MAPPING.items()):
        attr = vTodo.get(icalAttribute)
        if attr:
            # Convert ical type (vDates, vInt..) to sql alchemy understandable type (datetime, int...)
            attr = icalutils.convertIcalType(attr)
            if yokadiAttribute == "title":
                # Remove (id)
                attr = re.sub(r"\s?\(%s\)" % task.id, "", attr)
            if yokadiAttribute == "doneDate":
                # A done date defined indicate that task is done
                task.status = "done"
                # BUG: Done date is UTC, we must compute local time for yokadi
            if yokadiAttribute == "urgency":
                if attr == icalutils.yokadiUrgencyToIcalPriority(task.urgency):
                    # Priority does not change - don't update it
                    continue
                else:
                    # Priority has changed, we need to update urgency
                    attr = icalutils.icalPriorityToYokadiUrgency(int(attr))

            # Update attribute
            setattr(task, yokadiAttribute, attr)

    # Update keywords from categories
    if vTodo.get("categories"):
        if isinstance(vTodo.get("categories"), (list)):
            categories = vTodo.get("categories")
        else:
            categories = vTodo.get("categories").split(",")

        keywords = ["@%s" % k for k in categories]
        garbage, keywordFilters = parseutils.extractKeywords(" ".join(keywords))
        newKwDict = parseutils.keywordFiltersToDict(keywordFilters)
        if garbage:
            print("Got garbage while parsing categories: %s" % garbage)
        dbutils.createMissingKeywords(list(newKwDict.keys()), interactive=False)
        task.setKeywordDict(newKwDict)


class IcalHttpRequestHandler(http.server.BaseHTTPRequestHandler):
    """Simple Ical http request handler that only implement GET method"""
    newTask = {}  # Dict recording new task origin UID

    def do_GET(self):
        """Serve a GET request with complete todolist ignoring path"""
        self.send_response(200)
        self.end_headers()
        cal = generateCal()
        self.wfile.write(cal.to_ical())

    def do_PUT(self):
        """Receive a todolist for updating"""
        length = int(self.headers.getheader('content-length'))
        cal = icalendar.Calendar.from_ical(self.rfile.read(length))
        for vTodo in cal.walk():
            if "UID" in vTodo:
                try:
                    IcalHttpRequestHandler.processVTodo(self.newTask, vTodo)
                except YokadiException as e:
                    self.send_response(503, e)

        # Tell caller everything is ok
        self.send_response(200)
        self.end_headers()

    # This is static method to make it easier to test
    @staticmethod
    def processVTodo(newTaskDict, vTodo):
        session = db.getSession()
        uid = vTodo["UID"]
        if uid in newTaskDict:
            # This is a recent new task but remote ical calendar tool is not
            # aware of new Yokadi UID. Update it here to avoid duplicate new tasks
            print("update UID to avoid duplicate task")
            uid = TASK_UID % newTaskDict[uid]
            vTodo["UID"] = uid

        if uid.startswith(UID_PREFIX):
            # This is a yokadi Task.
            if vTodo["LAST-MODIFIED"].dt > vTodo["CREATED"].dt:
                # Task has been modified
                print("Modified task: %s" % uid)
                result = TASK_RE.match(uid)
                if result:
                    id = result.group(1)
                    task = dbutils.getTaskFromId(id)
                    print("Task found in yokadi db: %s" % task.title)
                    updateTaskFromVTodo(task, vTodo)
                    session.commit()
                else:
                    raise YokadiException("Task %s does exist in yokadi db " % uid)
        else:
            # This is a new task
            print("New task %s (%s)" % (vTodo["summary"], uid))
            keywordDict = {}
            task = dbutils.addTask(INBOX_PROJECT, vTodo["summary"],
                                   keywordDict, interactive=False)
            session.add(task)
            session.commit()
            # Keep record of new task origin UID to avoid duplicate
            # if user update it right after creation without reloading the
            # yokadi UID
            # TODO: add purge for old UID
            newTaskDict[uid] = task.id


class YokadiIcalServer(Thread):
    def __init__(self, port, listen):
        self.port = port
        if listen:
            self.address = ""
        else:
            self.address = "127.0.0.1"
        Thread.__init__(self)
        self.setDaemon(True)

    def run(self):
        """Method executed when the thread object start() method is called"""
        print("IcalServer starting...")
        icalServer = http.server.HTTPServer((self.address, self.port), IcalHttpRequestHandler)
        icalServer.serve_forever()
        print("IcalServer crash. Oups !")

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Project related commands.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.exc import IntegrityError

from yokadi.ycli import tui
from yokadi.ycli.completers import MultiCompleter, ProjectCompleter
from yokadi.ycli.basicparseutils import parseOneWordName
from yokadi.core import db
from yokadi.core.db import Project, Task
from yokadi.core.yokadiexception import YokadiException, BadUsageException
from yokadi.core.yokadioptionparser import YokadiOptionParser
from yokadi.core import dbutils


def getProjectFromName(name, parameterName="project_name"):
    """
    Helper function which returns a project given its name, or raise a
    YokadiException if it does not exist.
    """
    name = name.strip()
    if len(name) == 0:
        raise BadUsageException("Missing <%s> parameter" % parameterName)

    try:
        session = db.getSession()
        return session.query(Project).filter_by(name=name).one()
    except NoResultFound:
        raise YokadiException("Project '%s' not found. Use p_list to see all projects." % name)


class ProjectCmd(object):
    def do_p_add(self, line):
        """Add new project.
        p_add <projectName>"""
        if not line:
            print("Missing project name.")
            return
        projectName = parseOneWordName(line)
        session = db.getSession()
        try:
            project = Project(name=projectName)
            session.add(project)
            session.commit()
        except IntegrityError:
            session.rollback()
            raise YokadiException("A project named %s already exists. Please find another name" % projectName)
        print("Added project '%s'" % projectName)

    def do_p_edit(self, line):
        """Edit a project.
        p_edit <project name>"""
        session = db.getSession()
        project = dbutils.getOrCreateProject(line, createIfNeeded=False)

        if not project:
            raise YokadiException("Project does not exist.")

        # Edit
        line = tui.editLine(project.name)

        # Update project
        projectName = parseOneWordName(line)
        try:
            project.name = projectName
            session.commit()
        except IntegrityError:
            session.rollback()
            raise YokadiException("A project named %s already exists. Please find another name" % projectName)

    complete_p_edit = ProjectCompleter(1)

    def do_p_list(self, line):
        """List all projects."""
        session = db.getSession()
        for project in session.query(Project).all():
            if project.active:
                active = ""
            else:
                active = "(inactive)"
            taskCount = session.query(Task).filter_by(project=project).count()
            print("{:20} {:>4} {}".format(project.name, taskCount, active))

    def do_p_set_active(self, line):
        """Activate the given project"""
        session = db.getSession()
        project = getProjectFromName(line)
        project.active = True
        session.commit()
    complete_p_set_active = ProjectCompleter(1)

    def do_p_set_inactive(self, line):
        """Desactivate the given project"""
        session = db.getSession()
        project = getProjectFromName(line)
        project.active = False
        session.commit()
    complete_p_set_inactive = ProjectCompleter(1)

    def parser_p_remove(self):
        parser = YokadiOptionParser()
        parser.usage = "p_remove [options] <project_name>"
        parser.description = "Remove a project and all its associated tasks."
        parser.add_argument("-f", dest="force", default=False, action="store_true",
                            help="Skip confirmation prompt")
        parser.add_argument("project")
        return parser

    def do_p_remove(self, line):
        session = db.getSession()
        parser = self.parser_p_remove()
        args = parser.parse_args(line)
        project = getProjectFromName(args.project)
        nbTasks = len(project.tasks)
        if not args.force:
            if not tui.confirm("Remove project '%s' and its %d tasks" % (project.name, nbTasks)):
                return
        session.delete(project)
        session.commit()
        print("Project removed")
    complete_p_remove = ProjectCompleter(1)

    def parser_p_merge(self):
        parser = YokadiOptionParser()
        parser.usage = "p_remove <source_project> <destination_project>"
        parser.description = "Merge <source_project> into <destination_project>."
        parser.add_argument("source_project")
        parser.add_argument("destination_project")
        parser.add_argument("-f", dest="force", default=False, action="store_true",
                            help="Skip confirmation prompt")
        return parser

    def do_p_merge(self, line):
        session = db.getSession()
        parser = self.parser_p_merge()
        args = parser.parse_args(line)

        src = getProjectFromName(args.source_project)
        dst = getProjectFromName(args.destination_project)
        if not args.force:
            if not tui.confirm("Merge project '{}' into '{}'".format(src.name, dst.name)):
                return
        dst.merge(session, src)
        print("Project '{}' merged into '{}'".format(src.name, dst.name))
    complete_p_merge = MultiCompleter(ProjectCompleter(1), ProjectCompleter(2))

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Update from version 2 to version 3 of Yokadi DB

@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or newer
"""


def createProjectKeywordTable(cursor):
    cursor.execute("""
create table project_keyword (
    id integer not null,
    project_id integer,
    keyword_id integer,
    primary key (id),
    unique (project_id, keyword_id),
    foreign key(project_id) references project (id),
    foreign key(keyword_id) references keyword (id)
)
""")


def update(cursor):
    createProjectKeywordTable(cursor)


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



import unittest

from collections import namedtuple
from datetime import datetime

from dateutil import rrule

from yokadi.core.recurrencerule import RecurrenceRule
from yokadi.core.yokadiexception import YokadiException


# Use a date far away in the future because rrule does not work with dates in
# the past.
# This is a wednesday.
REF_DATE = datetime(2200, 3, 19, 21, 30)

TestRow = namedtuple("TestRow", ("text", "dct", "rule", "nextDate"))

TEST_DATA = [
    TestRow(
        "none",
        {},
        RecurrenceRule(),
        None,
    ),
    TestRow(
        "daily 17:15",
        {"freq": rrule.DAILY, "bymonth": (), "bymonthday": (), "byweekday": (), "byhour": (17,), "byminute": (15,)},
        RecurrenceRule(rrule.DAILY, byhour=17, byminute=15),
        REF_DATE.replace(day=20, hour=17, minute=15)
    ),
    TestRow(
        "weekly monday 10:00",
        {"freq": rrule.WEEKLY, "bymonth": (), "bymonthday": (), "byweekday": (0,), "byhour": (10,), "byminute": (0,)},
        RecurrenceRule(rrule.WEEKLY, byweekday=0, byhour=10),
        REF_DATE.replace(day=24, hour=10, minute=0)
    ),
    TestRow(
        "monthly 2 8:27",
        {"freq": rrule.MONTHLY, "bymonth": (), "bymonthday": (2,), "byweekday": (), "byhour": (8,), "byminute": (27,)},
        RecurrenceRule(rrule.MONTHLY, bymonthday=2, byhour=8, byminute=27),
        REF_DATE.replace(month=4, day=2, hour=8, minute=27)
    ),
    TestRow(
        "quarterly 2 8:27",
        {"freq": rrule.YEARLY, "bymonth": (1, 4, 7, 10), "bymonthday": (2,), "byweekday": (), "byhour": (8,),
         "byminute": (27,)},
        RecurrenceRule(rrule.YEARLY, bymonth=(1, 4, 7, 10), bymonthday=2, byhour=8, byminute=27),
        REF_DATE.replace(month=4, day=2, hour=8, minute=27)
    ),
    TestRow(
        "monthly first wednesday 8:27",
        {"freq": rrule.MONTHLY, "bymonth": (), "bymonthday": (), "byweekday": {"pos": 1, "weekday": 2}, "byhour": (8,),
         "byminute": (27,)},
        RecurrenceRule(rrule.MONTHLY, byweekday=RecurrenceRule.createWeekDay(pos=1, weekday=2), byhour=8, byminute=27),
        REF_DATE.replace(month=4, day=2, hour=8, minute=27)
    ),
    TestRow(
        "monthly last sunday 8:27",
        {"freq": rrule.MONTHLY, "bymonth": (), "bymonthday": (), "byweekday": {"pos": -1, "weekday": 6}, "byhour": (8,),
         "byminute": (27,)},
        RecurrenceRule(rrule.MONTHLY, byweekday=RecurrenceRule.createWeekDay(pos=-1, weekday=6), byhour=8, byminute=27),
        REF_DATE.replace(month=3, day=30, hour=8, minute=27)
    ),
    TestRow(
        "yearly 23/2 8:27",
        {"freq": rrule.YEARLY, "bymonth": (2,), "bymonthday": (23,), "byweekday": (), "byhour": (8,),
         "byminute": (27,)},
        RecurrenceRule(rrule.YEARLY, bymonth=2, bymonthday=23, byhour=8, byminute=27),
        REF_DATE.replace(year=2201, month=2, day=23, hour=8, minute=27)
    ),
]


class RecurrenceRuleTestCase(unittest.TestCase):
    def testFromHumaneString(self):
        testData = [
            ("daily 10:00", RecurrenceRule(rrule.DAILY, byhour=10)),
            ("weekly FR 23:00", RecurrenceRule(rrule.WEEKLY, byweekday=4, byhour=23)),
            ("none", RecurrenceRule()),
            ("weekly fr 23:00", RecurrenceRule(rrule.WEEKLY, byweekday=4, byhour=23)),
            ("weekly Fr 23:00", RecurrenceRule(rrule.WEEKLY, byweekday=4, byhour=23)),
            ("weekly Friday 23:00", RecurrenceRule(rrule.WEEKLY, byweekday=4, byhour=23)),
            ("monthly 3 13:00", RecurrenceRule(rrule.MONTHLY, bymonthday=3, byhour=13)),
            ("monthly second friday 13:00", RecurrenceRule(rrule.MONTHLY,
                                                           byweekday=RecurrenceRule.createWeekDay(weekday=4, pos=2),
                                                           byhour=13)),
            ("yearly 3/07 11:20", RecurrenceRule(rrule.YEARLY, bymonth=7, bymonthday=3, byhour=11, byminute=20)),
            ("quarterly 14 11:20", RecurrenceRule(rrule.YEARLY, bymonth=(1, 4, 7, 10), bymonthday=14, byhour=11,
                                                  byminute=20)),
            ("quarterly first monday 23:20", RecurrenceRule(rrule.YEARLY, bymonth=(1, 4, 7, 10),
                                                            byweekday=RecurrenceRule.createWeekDay(weekday=0, pos=1),
                                                            byhour=23, byminute=20)),
        ] + [(x.text, x.rule) for x in TEST_DATA]

        for text, expected in testData:
            with self.subTest(text=text):
                output = RecurrenceRule.fromHumaneString(text)
                self.assertEqual(output, expected,
                                 '\ninput:    {}\noutput:   {}\nexpected: {}'.format(text, output, expected))

    def testFromHumaneString_badInput(self):
        for badInput in ("foo",  # Unknown recurrence
                         "daily",  # No time
                         "weekly",  # No day
                         "weekly monday",  # No time
                         "monthly",  # No day
                         "monthly 10",  # No time
                         "quarterly",  # No day
                         "quarterly 10",  # No time
                         "monthly foo 12:00",  # Bad date
                         ):
            self.assertRaises(YokadiException, RecurrenceRule.fromHumaneString, badInput)

    def testToFromDict(self):
        for row in TEST_DATA:
            with self.subTest(text=row.text):
                rule = RecurrenceRule.fromDict(row.dct)
                self.assertEqual(rule, row.rule,
                                 '\ninput:    {}\nrule:     {}\nexpected: {}'.format(row.dct, rule, row.rule))
                dct = rule.toDict()
                self.assertEqual(dct, row.dct)

    def testGetNext(self):
        for row in TEST_DATA:
            with self.subTest(text=row.text):
                nextDate = row.rule.getNext(REF_DATE)
                self.assertEqual(nextDate, row.nextDate)



# -*- coding: UTF-8 -*-
"""
Simple rendering of t_list output

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""


class PlainListRenderer(object):
    def __init__(self, out):
        self.out = out
        self.first = True

    def addTaskList(self, sectionName, taskList):
        """Store tasks for this section
        @param sectionName: name of the task groupement section
        @type sectionName: unicode
        @param taskList: list of tasks to display
        @type taskList: list of db.Task instances
        """

        if not self.first:
            print(file=self.out)
        else:
            self.first = False
        print(sectionName, file=self.out)

        for task in taskList:
            print(("- " + task.title), file=self.out)

    def end(self):
        pass
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Standard codes for shell colors.

@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""

import sys

if sys.stdout.isatty():
    BOLD = '\033[01m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    ORANGE = '\033[33m'
    PURPLE = '\033[35m'
    CYAN = '\033[36m'
    GREY = '\033[37m'
    RESET = '\033[0;0m'
else:
    BOLD = ''
    RED = ''
    GREEN = ''
    ORANGE = ''
    PURPLE = ''
    CYAN = ''
    GREY = ''
    RESET = ''



"""
Date utilities.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
from datetime import datetime

from dateutil import rrule

from yokadi.core.ydateutils import getHourAndMinute, getWeekDayNumberFromDay, parseHumaneDateTime
from yokadi.core.yokadiexception import YokadiException


FREQUENCIES = {0: "Yearly", 1: "Monthly", 2: "Weekly", 3: "Daily"}

ALL_DAYS = (rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR, rrule.SA, rrule.SU)


class RecurrenceRule(object):
    """Thin wrapper around dateutil.rrule which brings:

    - Serialization to/from dict
    - Parsing methods
    - Sane defaults (byhour = byminute = bysecond = 0)
    - __eq__ operator
    - Readable name

    Dict format:
        freq: 0..3, see FREQUENCIES dict
        bymonth: tuple<1..12>
        bymonthday: tuple<1..31>
        byweekday: tuple<0..6> or {pos: -1;1..4, weekday: 0..6}
        byhour: tuple<0..23>
        byminute: tuple<0..59>

    Constructor arguments: same as dict format except tuples can be int or None
    for convenience
    """
    def __init__(self, freq=None, bymonth=None, bymonthday=None, byweekday=None, byhour=0, byminute=0):
        def tuplify(value):
            if value is None:
                return ()
            if isinstance(value, int):
                return (value,)
            else:
                return tuple(value)

        self._freq = freq
        self._bymonth = tuplify(bymonth)
        self._bymonthday = tuplify(bymonthday)
        if isinstance(byweekday, dict):
            self._byweekday = byweekday
        else:
            self._byweekday = tuplify(byweekday)
        self._byhour = tuplify(byhour)
        self._byminute = tuplify(byminute)

    @staticmethod
    def fromDict(dct):
        if not dct:
            return RecurrenceRule()
        return RecurrenceRule(**dct)

    @staticmethod
    def fromHumaneString(line):
        """Take a string following t_recurs format, returns a RecurrenceRule instance or None
        """
        freq = byminute = byhour = byweekday = bymonthday = bymonth = None

        tokens = line.split()

        tokens[0] = tokens[0].lower()

        if tokens[0] == "none":
            return RecurrenceRule()

        if tokens[0] == "daily":
            if len(tokens) != 2:
                raise YokadiException("You should give time for daily task")
            freq = rrule.DAILY
            byhour, byminute = getHourAndMinute(tokens[1])
        elif tokens[0] == "weekly":
            freq = rrule.WEEKLY
            if len(tokens) != 3:
                raise YokadiException("You should give day and time for weekly task")
            byweekday = getWeekDayNumberFromDay(tokens[1].lower())
            byhour, byminute = getHourAndMinute(tokens[2])
        elif tokens[0] in ("monthly", "quarterly"):
            if tokens[0] == "monthly":
                freq = rrule.MONTHLY
            else:
                # quarterly
                freq = rrule.YEARLY
                bymonth = (1, 4, 7, 10)
            if len(tokens) < 3:
                raise YokadiException("You should give day and time for %s task" % (tokens[0],))
            try:
                bymonthday = int(tokens[1])
                byhour, byminute = getHourAndMinute(tokens[2])
            except ValueError:
                POSITION = {"first": 1, "second": 2, "third": 3, "fourth": 4, "last": -1}
                if tokens[1].lower() in POSITION and len(tokens) == 4:
                    byweekday = RecurrenceRule.createWeekDay(
                        weekday=getWeekDayNumberFromDay(tokens[2].lower()),
                        pos=POSITION[tokens[1]])
                    byhour, byminute = getHourAndMinute(tokens[3])
                    bymonthday = None  # Default to current day number - need to be blanked
                else:
                    raise YokadiException("Unable to understand date. See help t_recurs for details")
        elif tokens[0] == "yearly":
            freq = rrule.YEARLY
            rDate = parseHumaneDateTime(" ".join(tokens[1:]))
            bymonth = rDate.month
            bymonthday = rDate.day
            byhour = rDate.hour
            byminute = rDate.minute
        else:
            raise YokadiException("Unknown frequency. Available: daily, weekly, monthly and yearly")

        return RecurrenceRule(
            freq,
            bymonth=bymonth,
            bymonthday=bymonthday,
            byweekday=byweekday,
            byhour=byhour,
            byminute=byminute,
        )

    def toDict(self):
        if not self:
            return {}

        return dict(
            freq=self._freq,
            bymonth=self._bymonth,
            bymonthday=self._bymonthday,
            byweekday=self._byweekday,
            byhour=self._byhour,
            byminute=self._byminute
        )

    def _rrule(self):
        if isinstance(self._byweekday, dict):
            day = ALL_DAYS[self._byweekday["weekday"]]
            byweekday = day(self._byweekday["pos"])
        else:
            byweekday = self._byweekday

        return rrule.rrule(
            freq=self._freq,
            bymonth=self._bymonth,
            bymonthday=self._bymonthday,
            byweekday=byweekday,
            byhour=self._byhour,
            byminute=self._byminute,
            bysecond=0
        )

    def getNext(self, refDate=None):
        """Return next date of recurrence after given date
        @param refDate: reference date used to compute the next occurence of recurrence
        @type refDate: datetime
        @return: next occurence (datetime)"""
        if not self:
            return None
        if refDate is None:
            refDate = datetime.now()
        refDate.replace(second=0, microsecond=0)
        return self._rrule().after(refDate)

    def getFrequencyAsString(self):
        """Return a string for the frequency"""
        if not self:
            return ""
        return FREQUENCIES[self._freq]

    @staticmethod
    def createWeekDay(pos, weekday):
        return dict(pos=pos, weekday=weekday)

    def __eq__(self, other):
        return self.toDict() == other.toDict()

    def __bool__(self):
        return self._freq is not None

    def __repr__(self):
        return repr(self.toDict())

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Task test cases
@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
import unittest
from unittest.mock import patch

import testutils

from yokadi.ycli import tui
from yokadi.ycli.main import YokadiCmd
from yokadi.core import db
from yokadi.core import dbutils
from yokadi.core.db import Task, TaskLock, Keyword, setDefaultConfig, Project, TaskKeyword
from yokadi.core.yokadiexception import YokadiException, BadUsageException


class TaskTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        setDefaultConfig()
        self.session = db.getSession()
        tui.clearInputAnswers()
        self.cmd = YokadiCmd()

    def testAdd(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")

        tui.addInputAnswers("y", "y")
        self.cmd.do_t_add("x @kw1 @kw2=12 t2")

        tui.addInputAnswers("n")
        self.cmd.do_t_add("notExistingProject newTask")

        tasks = self.session.query(Task).all()
        result = [x.title for x in tasks]
        expected = ["t1", "t2"]
        self.assertEqual(result, expected)

        kwDict = self.session.get(Task, 2).getKeywordDict()
        self.assertEqual(kwDict, dict(kw1=None, kw2=12))

        for bad_input in ("",  # No project
                          "x"):  # No task name
            self.assertRaises(BadUsageException, self.cmd.do_t_add, bad_input)

    def testEdit(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x txt @_note")

        tui.addInputAnswers("newtxt")
        self.cmd.do_t_edit("1")

        task = self.session.get(Task, 1)
        self.assertEqual(task.title, "newtxt")
        self.assertEqual(task.getKeywordDict(), {"_note": None})

    def testEditAddKeyword(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x txt")

        tui.addInputAnswers("txt @kw", "y")
        self.cmd.do_t_edit("1")

        task = self.session.get(Task, 1)
        self.assertEqual(task.title, "txt")
        self.assertEqual(task.getKeywordDict(), {"kw": None})

    def testEditRemoveKeyword(self):
        tui.addInputAnswers("y", "y")
        self.cmd.do_t_add("x txt @kw")

        tui.addInputAnswers("txt")
        self.cmd.do_t_edit("1")

        task = self.session.get(Task, 1)
        self.assertEqual(task.title, "txt")
        self.assertEqual(task.getKeywordDict(), {})

    def testRemove(self):
        # Create a recurrent task with one keyword
        tui.addInputAnswers("y", "y")
        self.cmd.do_t_add("x @kw bla")
        task = self.session.query(Task).one()
        self.cmd.do_t_recurs("1 daily 10:00")

        keyword = self.session.query(Keyword).filter_by(name="kw").one()
        self.assertEqual(keyword.tasks, [task])

        # Pretend we edit the task description so that we have a TaskLock for
        # this task
        taskLockManager = dbutils.TaskLockManager(task)
        taskLockManager.acquire()
        self.session.query(TaskLock).one()
        self.assertEqual(self.session.query(TaskLock).count(), 1)

        # Remove it, the keyword should no longer be associated with any task,
        # the lock should be gone
        tui.addInputAnswers("y")
        self.cmd.do_t_remove(str(task.id))

        self.assertEqual(keyword.tasks, [])
        self.assertEqual(self.session.query(TaskLock).count(), 0)

        # Should not crash
        taskLockManager.release()

    def testMark(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        task = self.session.get(Task, 1)
        self.assertEqual(task.status, "new")
        self.cmd.do_t_mark_started("1")
        self.assertEqual(task.status, "started")
        self.cmd.do_t_mark_new("1")
        self.assertEqual(task.status, "new")
        self.cmd.do_t_mark_done("1")
        self.assertEqual(task.status, "done")

    def testAddKeywords(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        task = self.session.get(Task, 1)

        tui.addInputAnswers("y", "y")
        self.cmd.do_t_add_keywords("1 @kw1 @kw2=12")

        kwDict = task.getKeywordDict()
        self.assertEqual(kwDict, dict(kw1=None, kw2=12))

        for bad_input in ("",  # No task
                          "1",  # No keyword
                          "1 kw1"):  # No @ before kw1
            self.assertRaises(YokadiException, self.cmd.do_t_add_keywords, bad_input)

    def testSetProject(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        tui.addInputAnswers("y")
        self.cmd.do_t_project("1 y")
        task1 = self.session.get(Task, 1)
        self.assertEqual(task1.project.name, "y")

        self.cmd.do_t_add("x t2")
        self.cmd.do_t_project("1 _")
        task1 = self.session.get(Task, 1)
        self.assertEqual(task1.project.name, "x")

        tui.addInputAnswers("n")
        self.cmd.do_t_project("1 doesnotexist")
        task1 = self.session.get(Task, 1)
        self.assertEqual(task1.project.name, "x")

    def testLastTaskId(self):
        # Using "_" with no prior task activity should raise an exception
        self.assertRaises(YokadiException, self.cmd.getTaskFromId, "_")

        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        task1 = self.session.get(Task, 1)
        self.assertEqual(self.cmd.getTaskFromId("_"), task1)

        self.cmd.do_t_add("x t2")
        task2 = self.session.get(Task, 2)
        self.assertEqual(self.cmd.getTaskFromId("_"), task2)

        self.cmd.do_t_mark_started("1")
        self.assertEqual(self.cmd.getTaskFromId("_"), task1)

    def testLastProjectName(self):
        # Using "_" with no prior project used should raise an exception
        self.assertRaises(YokadiException, self.cmd.do_t_add, "_ t1")
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        task1 = self.session.get(Task, 1)
        self.cmd.do_t_add("_ t2")
        task2 = self.session.get(Task, 2)
        self.assertEqual(task1.project, task2.project)

    def testRecurs(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        task = self.session.get(Task, 1)

        self.cmd.do_t_recurs("1 daily 10:00")
        self.assertTrue(task.recurrence)
        self.assertEqual(task.status, "new")
        self.cmd.do_t_mark_done("1")

        self.assertEqual(task.status, "new")

        self.cmd.do_t_recurs("1 none")
        self.assertFalse(task.recurrence)

        self.cmd.do_t_mark_done("1")
        self.assertEqual(task.status, "done")

        for bad_input in ("",  # No task
                          "1",  # No recurence
                          "1 foo",  # Unknown recurrence
                          "1 daily",  # No time
                          "1 weekly",  # No day
                          "1 weekly monday",  # No time
                          "1 monthly",  # No day
                          "1 monthly 10",  # No time
                          "1 quarterly",  # No day
                          "1 quarterly 10",  # No time
                          "1 monthly foo 12:00",  # Bad date
                          ):
            self.assertRaises(YokadiException, self.cmd.do_t_recurs, bad_input)

    def testRenderListSectionOrder(self):
        projectNames = "ccc", "aaa", "UPPER_CASE", "zzz", "mmm"
        projectList = []
        for name in projectNames:
            prj = Project(name=name)
            task = Task(project=prj, title="Hello")
            self.session.add(prj)
            self.session.add(task)
            projectList.append(prj)
        self.session.flush()

        renderer = testutils.TestRenderer()
        self.cmd._renderList(renderer, projectList, filters=[], order=[])

        self.assertEqual(list(renderer.taskDict.keys()), sorted(projectNames, key=lambda x: x.lower()))

    def testRenderListSectionOrderKeywords(self):
        prj = Project(name="prj")
        keywordNames = ["kw_" + x for x in ("ccc", "aaa", "UPPER_CASE", "zzz", "mmm")]
        keywordList = []
        for name in keywordNames:
            keyword = Keyword(name=name)
            task = Task(project=prj, title="Hello")
            TaskKeyword(task=task, keyword=keyword)
            self.session.add(task)
            keywordList.append(prj)
        self.session.flush()

        renderer = testutils.TestRenderer()
        self.cmd._renderList(renderer, [prj], filters=[], order=[], groupKeyword="kw_%")

        self.assertEqual(list(renderer.taskDict.keys()), sorted(keywordNames, key=lambda x: x.lower()))

    def testTlist(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        tui.addInputAnswers("y", "y")
        self.cmd.do_t_add("x @kw1 @kw2=12 t2")

        for line in ("", "-a", "-t", "-d today", "-u 10", "-k %", "-k _%", "-s t", "--overdue",
                     "@%", "@k%", "!@%", "!@kw1", "-f plain", "-f xml", "-f html", "-f csv"):
            self.cmd.do_t_list(line)

    def testTlistUrgency0(self):
        # Given a project with two tasks, one with a negative urgency
        prj = Project(name="prj")
        self.session.add(prj)
        t1 = Task(project=prj, title="t1")
        self.session.add(t1)
        t2 = Task(project=prj, title="t2", urgency=-1)
        self.session.add(t2)
        self.session.flush()
        # When I list tasks with -u 0
        renderer = testutils.TestRenderer()
        self.cmd.do_t_list("-u 0", renderer=renderer)
        # Then the task with a negative urgency is not listed
        self.assertEqual(renderer.tasks, [t1])

    def testNlist(self):
        tui.addInputAnswers("y")
        self.cmd.do_n_add("x t1")
        self.cmd.do_t_add("x t2")
        tui.addInputAnswers("y", "y")
        self.cmd.do_n_add("x @kw1 @kw2=12 t3")
        self.cmd.do_t_add("x @kw1 @kw2=12 t4")

        for line in ("", "-k %", "-k _%", "-s t",
                     "@%", "@k%", "!@%", "!@kw1", "-f plain"):
            self.cmd.do_t_list(line)

    def testTfilter(self):
        t1 = dbutils.addTask("x", "t1", interactive=False)
        t2 = dbutils.addTask("x", "t2", keywordDict={"kw1": None, "kw2": 12}, interactive=False)
        t3 = dbutils.addTask("y", "t3", interactive=False)

        testData = [
            ("@kw1", {"x": [t2]}),
            ("@kw1 @kw2", {"x": [t2]}),
            ("x", {"x": [t1, t2]}),
            ("x @kw1", {"x": [t2]}),
            ("none", {"x": [t1, t2], "y": [t3]}),
        ]
        for filter, expectedTaskDict in testData:
            self.cmd.do_t_filter(filter)
            renderer = testutils.TestRenderer()
            self.cmd.do_t_list("", renderer=renderer)
            self.assertEqual(renderer.taskDict.keys(), expectedTaskDict.keys())
            for key in renderer.taskDict.keys():
                self.assertEqual([x.title for x in renderer.taskDict[key]], [x.title for x in expectedTaskDict[key]])

        self.assertRaises(YokadiException, self.cmd.do_t_filter, "")

    def testTApply(self):
        self.cmd.do_k_add("lala")
        for i in range(10):
            tui.addInputAnswers("y")
            self.cmd.do_t_add("x t%s" % i)
        ids = [1, 2, 4, 5, 6, 9]
        self.cmd.do_t_apply("1 2,4-6 9 t_add_keywords @lala")
        for taskId in range(1, 10):
            kwDict = self.session.get(Task, taskId).getKeywordDict()
            if taskId in ids:
                self.assertEqual(kwDict, dict(lala=None))
            else:
                self.assertNotEqual(kwDict, dict(lala=None))

        # raise error if t_list had not been called previously
        self.assertRaises(BadUsageException, self.cmd.do_t_apply, "__ t_add_keywords @toto")

        self.cmd.do_t_list("@lala")
        self.cmd.do_t_apply("__ t_add_keywords @toto")
        for taskId in range(1, 10):
            kwDict = self.session.get(Task, taskId).getKeywordDict()
            if taskId in ids:
                self.assertEqual(kwDict, dict(lala=None, toto=None))
            else:
                self.assertNotEqual(kwDict, dict(lala=None, toto=None))

    def testReorderFailsOnInvalidInputs(self):
        self.assertRaises(BadUsageException, self.cmd.do_t_reorder, "unknown_project")
        self.assertRaises(BadUsageException, self.cmd.do_t_reorder, "too much args")

    def testDue(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")
        for valid_input in ("+1d", "+1m"):
            self.cmd.do_t_due("1 %s" % valid_input)
        for bad_input in ("coucou", "+1s"):
            self.assertRaises(YokadiException, self.cmd.do_t_due, "1 %s" % bad_input)

    def testToNote(self):
        tui.addInputAnswers("y")
        self.cmd.do_t_add("x t1")

        self.cmd.do_t_to_note(1)
        task = self.session.get(Task, 1)
        self.assertTrue(task.isNote(self.session))

        # Doing it twice should not fail
        self.cmd.do_t_to_note(1)
        task = self.session.get(Task, 1)
        self.assertTrue(task.isNote(self.session))

    def testToTask(self):
        tui.addInputAnswers("y")
        self.cmd.do_n_add("x t1")

        self.cmd.do_n_to_task(1)
        task = self.session.get(Task, 1)
        self.assertFalse(task.isNote(self.session))

        # Doing it twice should not fail
        self.cmd.do_n_to_task(1)
        task = self.session.get(Task, 1)
        self.assertFalse(task.isNote(self.session))

    @patch("yokadi.ycli.tui.editText")
    def testReorder(self, editTextMock):
        t1, t2, t3 = [dbutils.addTask("x", f"t{x}", interactive=False) for x in range(1, 4)]

        # Simulate moving t3 from 3rd line to the 1st
        editTextMock.return_value = "3,t3\n1,t1\n2,t2"

        self.cmd.do_t_reorder("x")
        editTextMock.assert_called_with("1,t1\n2,t2\n3,t3")

        self.assertEqual(t3.urgency, 2)
        self.assertEqual(t1.urgency, 1)
        self.assertEqual(t2.urgency, 0)

# vi: ts=4 sw=4 et



"""
Update from version 11 to version 12 of Yokadi DB

- Decrypt all encrypted tasks

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""
import base64

from getpass import getpass

from yokadi.update import updateutils
from yokadi.ycli import tui

try:
    from Crypto.Cipher import AES as CRYPTO_ALGO
except ImportError:
    CRYPTO_ALGO = None

CRYPTO_PREFIX = "---YOKADI-ENCRYPTED-MESSAGE---"
KEY_LENGTH = 32

CRYPTO_CHECK_KEY = "CRYPTO_CHECK"
CONFIG_KEYS = "PASSPHRASE_CACHE", CRYPTO_CHECK_KEY


def getPassphrase():
    phrase = getpass(prompt="Enter passphrase: ")
    phrase = phrase[:KEY_LENGTH]
    return phrase.ljust(KEY_LENGTH, " ")


def decryptData(cypher, data):
    if not data:
        return data
    data = data[len(CRYPTO_PREFIX):]  # Remove crypto prefix
    data = base64.b64decode(data)
    return cypher.decrypt(data).rstrip().decode(encoding="utf-8")


def decryptTask(cursor, cypher, row):
    taskId, title, description = row
    title = decryptData(cypher, title)
    description = decryptData(cypher, description)
    cursor.execute("update task set title = ?, description = ? where id = ?",
                   (title, description, taskId))


def getCheckText(cursor):
    sql = "select value from config where name like ?"
    row = cursor.execute(sql, (CRYPTO_CHECK_KEY,)).fetchone()
    return row[0]


def checkPassphrase(cypher, checkText):
    try:
        decryptData(cypher, checkText)
        return True
    except UnicodeDecodeError:
        return False


def decryptEncryptedTasks(cursor):
    sql = "select id, title, description from task where title like ?"

    rows = cursor.execute(sql, (CRYPTO_PREFIX + "%",)).fetchall()
    if not rows:
        return

    if CRYPTO_ALGO is None:
        msg = ("This database contains encrypted data but pycrypto is not"
               " installed.\n"
               "Please install pycrypto and try again.")
        raise updateutils.UpdateError(msg)

    if not tui.confirm("This database contains encrypted tasks, but Yokadi no "
                       "longer supports encryption.\n"
                       "These tasks need to be decrypted to continue using "
                       "Yokadi.\n"
                       "Do you want to decrypt your tasks?"):
        raise updateutils.UpdateCanceledError()

    checkText = getCheckText(cursor)
    while True:
        phrase = getPassphrase()
        cypher = CRYPTO_ALGO.new(phrase)
        if checkPassphrase(cypher, checkText):
            break
        else:
            if not tui.confirm("Wrong passphrase, try again?"):
                raise updateutils.UpdateCanceledError()
    for row in rows:
        decryptTask(cursor, cypher, row)


def removeCryptoConfigKeys(cursor):
    sql = "delete from config where name like ?"
    for key in CONFIG_KEYS:
        cursor.execute(sql, (key,))


def update(cursor):
    decryptEncryptedTasks(cursor)
    removeCryptoConfigKeys(cursor)


if __name__ == "__main__":
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Date utilities test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""

import unittest

from datetime import datetime

from yokadi.core import dbutils, db
from yokadi.ycli import tui
from yokadi.core.db import Keyword, Project
from yokadi.core.yokadiexception import YokadiException


class DbUtilsTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        tui.clearInputAnswers()

    def testGetTaskFromId(self):
        tui.addInputAnswers("y")
        t1 = dbutils.addTask("x", "t1", {})

        task = dbutils.getTaskFromId(str(t1.id))
        self.assertEqual(task, t1)

        task = dbutils.getTaskFromId(t1.id)
        self.assertEqual(task, t1)

        task = dbutils.getTaskFromId(t1.uuid)
        self.assertEqual(task, t1)

    def testGetOrCreateKeyword(self):
        # interactive
        tui.addInputAnswers("y")
        dbutils.getOrCreateKeyword("k1")
        self.session.query(Keyword).filter_by(name="k1").one()

        # !interactive
        dbutils.getOrCreateKeyword("k2", interactive=False)
        self.session.query(Keyword).filter_by(name="k2").one()

    def testGetOrCreateProject(self):
        # interactive
        tui.addInputAnswers("y")
        dbutils.getOrCreateProject("p1")
        self.session.query(Project).filter_by(name="p1").one()

        # !interactive
        dbutils.getOrCreateProject("p2", interactive=False)
        self.session.query(Project).filter_by(name="p2").one()

    def testGetKeywordFromName(self):
        tui.addInputAnswers("y")
        k1 = dbutils.getOrCreateKeyword("k1", self.session)
        self.assertRaises(YokadiException, dbutils.getKeywordFromName, "")
        self.assertRaises(YokadiException, dbutils.getKeywordFromName, "foo")
        self.assertEqual(k1, dbutils.getKeywordFromName("k1"))

    def testTaskLockManagerStaleLock(self):
        tui.addInputAnswers("y")
        t1 = dbutils.addTask("x", "t1", {})
        taskLockManager = dbutils.TaskLockManager(t1)

        # Lock the task
        taskLockManager.acquire(pid=1, now=datetime(2014, 1, 1))
        lock1 = taskLockManager._getLock()
        self.assertEqual(lock1.pid, 1)

        # Try to lock again, the stale lock should get reused
        taskLockManager.acquire(pid=2, now=datetime(2015, 1, 1))
        lock2 = taskLockManager._getLock()
        self.assertEqual(lock1.id, lock2.id)
        self.assertEqual(lock2.pid, 2)

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Xml rendering of t_list output

@author: Aurélien Gâteau <mail@agateau.com>
@author: Sébastien Renard <sebastien.renard@digitalfox.org>
@license: GPL v3 or later
"""
from xml.dom import minidom as dom

TASK_FIELDS = ["title", "creationDate", "dueDate", "doneDate", "description", "urgency", "status", "keywords"]


class XmlListRenderer(object):
    def __init__(self, out):
        self.out = out
        self.doc = dom.Document()
        self.rootElement = self.doc.createElement("yokadi")
        self.doc.appendChild(self.rootElement)

    def addTaskList(self, sectionName, taskList):
        """Store tasks for this section
        @param sectionName: name of the task groupement section
        @type sectionName: unicode
        @param taskList: list of tasks to display
        @type taskList: list of db.Task instances
        """

        sectionElement = self.doc.createElement("section")
        self.rootElement.appendChild(sectionElement)
        sectionElement.setAttribute("name", sectionName)

        for task in taskList:
            taskElement = self.doc.createElement("task")
            sectionElement.appendChild(taskElement)

            taskElement.setAttribute("id", str(task.id))

            for field in TASK_FIELDS:
                if field == "keywords":
                    self._exportKeywords(taskElement, task.getKeywordDict())
                elif field == "description":
                    if task.description:
                        descriptionElement = self.doc.createElement("description")
                        taskElement.appendChild(descriptionElement)
                        descriptionElement.appendChild(self.doc.createTextNode(task.description))
                else:
                    taskElement.setAttribute(field, str(task.__getattribute__(field)))

    def _exportKeywords(self, taskElement, keywordDict):
        for key, value in list(keywordDict.items()):
            keywordElement = self.doc.createElement("keyword")
            taskElement.appendChild(keywordElement)
            keywordElement.setAttribute("name", str(key))
            if value:
                keywordElement.setAttribute("value", str(value))

    def end(self):
        self.out.write(self.doc.toprettyxml(indent="    "))
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Date utilities test cases
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""

import unittest
import operator
from datetime import datetime, timedelta, time

from yokadi.core import ydateutils
from yokadi.core.yokadiexception import YokadiException


class YDateUtilsTestCase(unittest.TestCase):
    def testParseDateTimeDelta(self):
        testData = [
            ("1m", timedelta(minutes=1)),
            ("3M", timedelta(minutes=3)),
            ("5H", timedelta(hours=5)),
            ("6.5D", timedelta(days=6, hours=12)),
            ("12W", timedelta(days=12 * 7)),
        ]

        for text, expected in testData:
            output = ydateutils.parseDateTimeDelta(text)
            self.assertEqual(expected, output)

    def testParseHumaneDateTime(self):
        for date in ("+5M", "+1m", "+2H", "+3h", "+9D", "+14d", "+432W", "+0w",
                     "01/01/2009", "10/10/2008 12", "7/7/2007 10:15", "1/2/2003 1:2:3"):
            ydateutils.parseHumaneDateTime(date)

        for invalidDate in ("2008", "01/2009", "01//02/01", "02/20/2009", "", "+3e", "lkjljlkjlkj", "200/200/2009"):
            self.assertRaises(YokadiException, ydateutils.parseHumaneDateTime, invalidDate)

        # Fake today to a fixed date. This is a saturday (weekday=5).
        today = datetime(2009, 1, 3)
        endOfDay = dict(hour=23, minute=59, second=59)
        startOfDay = dict(hour=0, minute=0, second=0)
        testData = [
            ("06/02/2009", None, datetime(2009, 2, 6)),
            ("06/02/09", None, datetime(2009, 2, 6)),
            ("06/02/2009 12:30", None, datetime(2009, 2, 6, 12, 30)),
            ("06/02/2009", ydateutils.TIME_HINT_BEGIN, datetime(2009, 2, 6, 0, 0, 0)),
            ("06/02/2009", ydateutils.TIME_HINT_END, datetime(2009, 2, 6, 23, 59, 59)),
            ("tomorrow 18:00", None, today + timedelta(days=1, hours=18)),
            ("tomorrow", ydateutils.TIME_HINT_END, today.replace(day=4, hour=23, minute=59, second=59)),
            ("sunday", None, datetime(2009, 1, 4)),
            ("tu 11:45", None, datetime(2009, 1, 6, 11, 45)),
            ("today", ydateutils.TIME_HINT_END, today.replace(**endOfDay)),
            ("today", ydateutils.TIME_HINT_BEGIN, today.replace(**startOfDay)),
            ("now", None, today),
            ("today 5 PM", None, today.replace(hour=17)),
            ("6:32pm", None, today.replace(hour=18, minute=32)),
            ("06/02 2:40 am", None, datetime(2009, 2, 6, 2, 40)),
            ("+2w", None, datetime(2009, 1, 17)),
            ("+1d", None, datetime(2009, 1, 4)),
            ("-1d", None, datetime(2009, 1, 2)),
            ("+3h", None, datetime(2009, 1, 3, 3, 0)),
            ("-1M", None, datetime(2009, 1, 2, 23, 59)),
        ]

        for text, hint, expected in testData:
            output = ydateutils.parseHumaneDateTime(text, hint=hint, today=today)
            self.assertEqual(expected, output)

    def testPastTime(self):
        # Fake now to a fixed date and time
        now = datetime(2009, 1, 3, 11, 0)
        testData = [
            ("1:00pm", now.replace(hour=13, minute=0)),
            ("10:00am", now.replace(hour=10, minute=0) + timedelta(days=1)),
        ]

        for text, expected in testData:
            output = ydateutils.parseHumaneDateTime(text, hint=None, today=now)
            self.assertEqual(expected, output)

    def testFormatTimeDelta(self):
        testData = [
            (timedelta(minutes=1), "1m"),
            (timedelta(hours=1), "1h 0m"),
            (timedelta(hours=1, minutes=1), "1h 1m"),
            (timedelta(days=2, hours=5), "2d"),
            (timedelta(days=7), "1w"),
            (timedelta(days=10), "1w, 3d"),
            (timedelta(days=60), "2M"),
            (timedelta(days=80), "2M, 20d"),
            (timedelta(days=365), "1Y"),
            (timedelta(days=400), "1Y, 1M"),
        ]

        for input, expected in testData:
            output = ydateutils.formatTimeDelta(input)
            self.assertEqual(expected, output)
            output = ydateutils.formatTimeDelta(-input)
            self.assertEqual("-" + expected, output)

    def testParseDateLimit(self):
        # Fake today to a fixed date. This is a saturday (weekday=5).
        today = datetime(2009, 1, 3)
        endOfDay = dict(hour=23, minute=59, second=59)
        startOfDay = dict(hour=0, minute=0, second=0)

        testData = [
            ("today", operator.__le__, today.replace(**endOfDay)),
            ("<=today", operator.__le__, today.replace(**endOfDay)),
            ("<today", operator.__lt__, today.replace(**startOfDay)),
            (">today", operator.__gt__, today.replace(**endOfDay)),
            (">=today", operator.__ge__, today.replace(**startOfDay)),
            ("<=06/02/2009", operator.__le__, datetime(2009, 2, 6).replace(**endOfDay)),
            ("<06/02/2009", operator.__lt__, datetime(2009, 2, 6).replace(**startOfDay)),
            ("tomorrow 18:00", operator.__le__, today + timedelta(days=1, hours=18)),
            ("sunday", operator.__le__, datetime(2009, 1, 4).replace(**endOfDay)),
            ("tu 11:45", operator.__le__, datetime(2009, 1, 6, 11, 45)),
        ]

        for text, expectedOp, expectedDate in testData:
            output = ydateutils.parseDateLimit(text, today=today)
            output = ydateutils.parseDateLimit(text, today=today)
            self.assertEqual(expectedOp, output[0])
            self.assertEqual(expectedDate, output[1])

    def testGuessTime(self):
        for invalidTime in ("+5M", "+1m", "+2H", "+3h", "+9D", "+14d", "+432W", "+0w",
                            "01/01/2009", "10/10/2008 12", "7/7/2007 10:15", "1/2/2003 1:2:3"):
            self.assertIsNone(ydateutils.guessTime(invalidTime))

        for text, expected in (('12:05:20', time(hour=12, minute=5, second=20)), ('10:00am', time(hour=10)),
                               ('7:30pm', time(hour=19, minute=30))):
            output = ydateutils.guessTime(text)
            self.assertEqual(expected, output)

# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Alias test cases
@author: Benjamin Port <benjamin.port@ben2367.fr>
@license: GPL v3 or later
"""

import unittest
import sys
from io import StringIO

from yokadi.core import db
from yokadi.core.db import setDefaultConfig
from yokadi.core.yokadiexception import YokadiException
from yokadi.ycli import tui
from yokadi.ycli.main import YokadiCmd


class ConfTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        setDefaultConfig()
        self.session = db.getSession()
        tui.clearInputAnswers()
        self.cmd = YokadiCmd()

    def testConfig(self):
        out = StringIO()
        oldstdout = sys.stdout
        tui.stdout = out
        self.cmd.do_c_set("ALARM_DELAY 69")
        self.cmd.do_c_get("ALARM_DELAY")
        self.assertTrue("ALARM_DELAY" in out.getvalue())
        self.assertTrue("69" in out.getvalue())
        tui.stdout = oldstdout

    def testPositiveValueConfig(self):
        self.assertRaises(YokadiException, self.cmd.do_c_set, "ALARM_DELAY -1")
        self.assertRaises(YokadiException, self.cmd.do_c_set, "ALARM_SUSPEND -1")
        self.assertRaises(YokadiException, self.cmd.do_c_set, "PURGE_DELAY -1")

    def testWrongKey(self):
        self.assertRaises(YokadiException, self.cmd.do_c_set, "BAD_KEY value")
        self.assertRaises(YokadiException, self.cmd.do_c_get, "BAD_KEY")



# -*- coding: UTF-8 -*-
"""
Basepaths test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import os
import shutil
import tempfile

from pathlib import Path

from yokadi.core import basepaths
from yokadi.tests.yokaditestcase import YokadiTestCase


class BasePathsUnixTestCase(YokadiTestCase):
    def setUp(self):
        YokadiTestCase.setUp(self)
        self._oldWindows = basepaths._WINDOWS
        basepaths._WINDOWS = False

    def tearDown(self):
        basepaths._WINDOWS = self._oldWindows
        YokadiTestCase.tearDown(self)

    def testMigrateOldDb(self):
        oldDb = Path(self.testHomeDir) / '.yokadi.db'
        newDb = Path(basepaths.getDbPath(basepaths.getDataDir()))

        oldDb.touch()

        basepaths.migrateOldDb(str(newDb))
        self.assertFalse(oldDb.exists())
        self.assertTrue(newDb.exists())

    def testMigrateNothingToDo(self):
        newDb = Path(basepaths.getDbPath(basepaths.getDataDir()))
        basepaths.migrateOldDb(str(newDb))
        basepaths.migrateOldHistory()
        self.assertFalse(newDb.exists())

    def testMigrateOldDbFails(self):
        oldDb = Path(self.testHomeDir) / '.yokadi.db'
        newDb = Path(basepaths.getDbPath(basepaths.getDataDir()))

        oldDb.touch()
        newDb.parent.mkdir(parents=True)
        newDb.touch()

        self.assertRaises(basepaths.MigrationException, basepaths.migrateOldDb, str(newDb))

    def testMigrateOldHistory(self):
        old = Path(self.testHomeDir) / '.yokadi_history'
        new = Path(basepaths.getHistoryPath())

        old.touch()

        basepaths.migrateOldHistory()
        self.assertFalse(old.exists())
        self.assertTrue(new.exists())

    def testMigrateOldHistoryOverwriteNew(self):
        old = Path(self.testHomeDir) / '.yokadi_history'
        new = Path(basepaths.getHistoryPath())

        with old.open('w') as f:
            f.write('old')
        new.parent.mkdir(parents=True)
        with new.open('w') as f:
            f.write('new')

        basepaths.migrateOldHistory()
        self.assertFalse(old.exists())
        with new.open() as f:
            newData = f.read()
        self.assertEqual(newData, 'old')

    def testHistoryEnvVar(self):
        path = "foo"
        os.environ["YOKADI_HISTORY"] = path
        self.assertEqual(basepaths.getHistoryPath(), path)

    def testDbEnvVar(self):
        path = "foo"
        os.environ["YOKADI_DB"] = path
        self.assertEqual(basepaths.getDbPath(basepaths.getDataDir()), path)


class BasePathsWindowsTestCase(YokadiTestCase):
    def setUp(self):
        YokadiTestCase.setUp(self)
        self._oldWindows = basepaths._WINDOWS
        basepaths._WINDOWS = True
        self.testAppDataDir = tempfile.mkdtemp(prefix="yokadi-basepaths-testcase")
        os.environ["APPDATA"] = self.testAppDataDir

    def tearDown(self):
        shutil.rmtree(self.testAppDataDir)
        basepaths._WINDOWS = self._oldWindows
        YokadiTestCase.tearDown(self)

    def testGetCacheDir(self):
        expected = os.path.join(self.testAppDataDir, "yokadi", "cache")
        self.assertEqual(basepaths.getCacheDir(), expected)

    def testGetDataDir(self):
        expected = os.path.join(self.testAppDataDir, "yokadi", "data")
        self.assertEqual(basepaths.getDataDir(), expected)

    def testOldHistoryPath(self):
        expected = os.path.join(self.testAppDataDir, ".yokadi_history")
        self.assertEqual(basepaths._getOldHistoryPath(), expected)



# coding:utf-8
"""
Ical features test cases
@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""


import unittest
from datetime import datetime, timedelta

import icalendar

from yokadi.ycli import tui
from yokadi.ycli.projectcmd import getProjectFromName
from yokadi.yical import yical
from yokadi.core import dbutils
from yokadi.core import db


class IcalTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        tui.clearInputAnswers()

    def testUrgencyMapping(self):
        tui.addInputAnswers("y")
        t1 = dbutils.addTask("x", "t1", {})
        v1 = yical.createVTodoFromTask(t1)

        self.assertEqual(v1.get("priority"), None)  # No priority

        t1.urgency = 45
        v1 = yical.createVTodoFromTask(t1)
        self.assertEqual(v1.get("priority"), 2)

        yical.updateTaskFromVTodo(t1, v1)
        self.assertEqual(t1.urgency, 45)  # Ensure urgency does change

        v1["priority"] = 4
        yical.updateTaskFromVTodo(t1, v1)
        self.assertEqual(t1.urgency, 20)  # Check urgency is updated

    def testTitleMapping(self):
        tui.addInputAnswers("y")
        t1 = dbutils.addTask("x", "t1", {})
        v1 = yical.createVTodoFromTask(t1)

        # Check id is here
        self.assertEqual(v1.get("summary")[-4:], " (%s)" % t1.id)

        # Title and id should not change with update
        origin_id = t1.id
        origin_title = t1.title
        yical.updateTaskFromVTodo(t1, v1)
        self.assertEqual(t1.id, origin_id)
        self.assertEqual(t1.title, origin_title)

        # Update vtodo summary and remove (id) or space before (id) info.
        # Only task title should be changed
        for new_summary in ("hello", "hello(%s)" % origin_id, "hello (%s)" % origin_id,
                            "(%s)hello" % origin_id, " (%s)hello" % origin_id):
            v1["summary"] = new_summary
            yical.updateTaskFromVTodo(t1, v1)
            self.assertEqual(t1.id, origin_id)
            self.assertEqual(t1.title, "hello")

        # Update votod with fake id info.
        # Should be present in task title
        for new_summary in ("hello", "hello()", "hello(123456)", "hello (123456)"):
            v1["summary"] = new_summary
            yical.updateTaskFromVTodo(t1, v1)
            self.assertEqual(t1.id, origin_id)
            self.assertEqual(t1.title, new_summary)

    def testKeywordMapping(self):
        tui.addInputAnswers("y")
        tui.addInputAnswers("y")
        tui.addInputAnswers("y")
        t1 = dbutils.addTask("x", "t1", {"k1": None, "k2": 123})
        v1 = yical.createVTodoFromTask(t1)

        # Check categories are created
        categories = [str(c) for c in v1.get("categories")]
        categories.sort()
        self.assertEqual(categories, ["k1", "k2=123"])

        # Check keywords are still here
        yical.updateTaskFromVTodo(t1, v1)
        keywords = list(t1.getKeywordDict().keys())
        self.session.commit()
        keywords.sort()
        self.assertEqual(keywords, ["k1", "k2"])
        self.assertEqual(t1.getKeywordDict()["k2"], 123)

        # Remove k2 category
        v1["categories"] = ["k1"]
        yical.updateTaskFromVTodo(t1, v1)
        self.session.commit()
        self.assertEqual(list(t1.getKeywordDict().keys()), ["k1", ])

        # Set k1 value
        v1["categories"] = ["k1=456", ]
        yical.updateTaskFromVTodo(t1, v1)
        self.session.commit()
        self.assertEqual(t1.getKeywordDict()["k1"], 456)

        # Create a category
        v1["categories"] = ["k1", "k4=789"]
        yical.updateTaskFromVTodo(t1, v1)
        keywords = list(t1.getKeywordDict().keys())
        keywords.sort()
        self.assertEqual(keywords, ["k1", "k4"])
        self.assertEqual(t1.getKeywordDict()["k4"], 789)

    def testTaskDoneMapping(self):
        tui.addInputAnswers("y")
        t1 = dbutils.addTask("x", "t1", {})
        v1 = yical.createVTodoFromTask(t1)

        completed = datetime.now()
        v1.add("COMPLETED", completed)
        yical.updateTaskFromVTodo(t1, v1)
        self.assertEqual(t1.status, "done")
        self.assertEqual(t1.doneDate, completed)

    def testGenerateCal(self):
        # Add an inactive project
        t1 = dbutils.addTask("p1", "t1", interactive=False)
        project = getProjectFromName("p1")
        project.active = False

        # And an active project with 3 tasks, one of them is done
        t2new = dbutils.addTask("p2", "t2new", interactive=False)

        t2started = dbutils.addTask("p2", "t2started", interactive=False)
        t2started.setStatus("started")

        t2done = dbutils.addTask("p2", "t2done", interactive=False)
        t2done.setStatus("done")

        self.session.commit()

        # Generate the calendar
        cal = yical.generateCal()

        # It should contain only "p2", "t1" and "t2new" and "t2started"
        # I am not sure that it should contain "t1" (since its project is not active), but that's the current behavior
        summaries = sorted(str(x["SUMMARY"]) for x in cal.subcomponents)
        expected = sorted(["p2", f"t1 ({t1.id})", f"t2new ({t2new.id})", f"t2started ({t2started.id})"])

        self.assertEqual(summaries, expected)

    def testHandlerProcessVTodoModifyTask(self):
        # Create a task
        task = dbutils.addTask("p1", "t1", interactive=False)
        self.session.commit()

        # Create a vTodo to modify the task
        modified = datetime.now()
        created = modified + timedelta(hours=-1)
        vTodo = icalendar.Todo()
        vTodo["UID"] = yical.TASK_UID % str(task.id)
        vTodo.add("CREATED", created)
        vTodo.add("LAST-MODIFIED", modified)
        vTodo.add("summary", "new title")

        # Process the vTodo
        newTaskDict = {}
        yical.IcalHttpRequestHandler.processVTodo(newTaskDict, vTodo)

        # The task title must have changed
        task = dbutils.getTaskFromId(task.id)
        self.assertEqual(task.title, "new title")

        # newTaskDict must not have changed
        self.assertEqual(newTaskDict, {})

    def testHandlerProcessVTodoCreateTask(self):
        # Create a vTodo to add a new task
        vTodo = icalendar.Todo()
        vTodo["UID"] = "zogzog"
        vTodo.add("summary", "new task")

        # Process the vTodo
        newTaskDict = {}
        yical.IcalHttpRequestHandler.processVTodo(newTaskDict, vTodo)

        # The task should be in newTaskDict
        newTaskList = list(newTaskDict.items())
        self.assertEqual(len(newTaskList), 1)

        (uid, taskId) = newTaskList[0]

        # And the task can be retrieved
        task = dbutils.getTaskFromId(taskId)
        self.assertEqual(task.title, "new task")

        # And there is only one task
        self.assertEqual(self.session.query(db.Task).count(), 1)



# -*- coding: UTF-8 -*-
"""
Database utilities.

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
from collections import namedtuple
from datetime import datetime

from sqlalchemy import desc

from yokadi.core import db
from yokadi.core.db import Task, NOTE_KEYWORD
from yokadi.core.yokadiexception import YokadiException
from yokadi.core import dbutils
from yokadi.core.dbutils import KeywordFilter
from yokadi.ycli import parseutils


MEditEntry = namedtuple("MEditEntry", ["id", "status", "title", "keywords"])


DOC_COMMENT = """
Line format: <id> <status> <task title>

To change the title of a task, just change the text after the status character.
You can add or remove keywords like you do when using t_add.

To change the status of a task, change the status character to one of:
 N new
 S started
 D done

To add a new task, add a new line using '-' for the task id. If you do not
specify a status, the task will be marked as new. Examples:

- This is a new task
- N This is another new task
- S This task has already been started

To adjust task urgencies, re-order the lines.

To remove a task, delete its line or comment it out with '#'.

To cancel changes, quit without saving.

Empty lines and lines starting with a '#' are ignored.

Warning: Do NOT edit the task id, this will confuse Yokadi.
"""


class ParseError(YokadiException):
    def __init__(self, message, lineNumber, line):
        fullMessage = "Error line %d (\"%s\"): %s" % (lineNumber + 1, line, message)
        YokadiException.__init__(self, fullMessage)
        self.lineNumber = lineNumber
        self.message = message


def createMEditText(entries, docComment=DOC_COMMENT):
    def formatLine(entry):
        status = entry.status[0].upper()
        line = parseutils.createLine(None, entry.title, entry.keywords)
        return "%d %s %s" % (entry.id, status, line)

    def prefixComment(comment):
        for line in comment.strip().splitlines():
            if line:
                yield "# " + line
            else:
                yield "#"

    lines = [formatLine(x) for x in entries]
    lines.append('')
    lines.extend(prefixComment(docComment))
    return "\n".join(lines) + "\n"


def parseMEditLine(line):
    tokens = line.split(" ", 2)
    nbTokens = len(tokens)
    if nbTokens < 3:
        if nbTokens == 2 and tokens[0] == "-":
            # Special case: adding a one-word new task
            tokens.append("")
        else:
            raise Exception("Invalid line")

    if tokens[0] == "-":
        id = None
    else:
        try:
            id = int(tokens[0])
        except ValueError:
            raise Exception("Invalid id value")

    statusChar = tokens[1].lower()
    line = tokens[2]
    if statusChar == "n":
        status = "new"
    elif statusChar == "s":
        status = "started"
    elif statusChar == "d":
        status = "done"
    elif id is None:
        # Special case: if this is a new task, then statusChar is actually a
        # one-letter word starting the task title
        status = "new"
        line = tokens[1] + ((" " + line) if line else "")
    else:
        raise Exception("Invalid status")

    _, title, keywords = parseutils.parseLine("dummy " + line)
    return MEditEntry(id, status, title, keywords)


def parseMEditText(text):
    lst = []
    ids = set()
    for num, line in enumerate(text.split("\n")):
        line = line.strip()
        if not line or line[0] == "#":
            continue

        try:
            entry = parseMEditLine(line)
        except Exception as exc:
            exc = ParseError(str(exc), lineNumber=num + 1, line=line)
            raise exc

        if entry.id is not None:
            if entry.id in ids:
                exc = ParseError("Duplicate id value", lineNumber=num + 1, line=line)
                raise exc
            ids.add(entry.id)

        lst.append(entry)
    return lst


def createEntriesForProject(project):
    session = db.getSession()
    lst = session.query(Task).filter(Task.projectId == project.id,
                                     Task.status != 'done')

    lst = KeywordFilter(NOTE_KEYWORD, negative=True).apply(lst)
    lst = lst.order_by(desc(Task.urgency))
    return [createEntryForTask(x) for x in lst]


def createEntryForTask(task):
    return MEditEntry(task.id, task.status, task.title, task.getKeywordDict())


def applyChanges(project, oldList, newList, interactive=True):
    """
    Modify a project so that its task list is newList

    @param project: the project name
    @param oldList: a list of Task instances
    @param newList: a list of MEditEntry
    @param interactive: whether to confirm creation of new keywords
    """
    session = db.getSession()

    # Sanity check: all ids in newList should be in oldList
    oldIds = set([x.id for x in oldList])
    newIds = set([x.id for x in newList if x.id is not None])
    unknownIds = newIds.difference(oldIds)
    if unknownIds:
        idString = ", ".join([str(x) for x in unknownIds])
        raise YokadiException("Unknown id(s): %s" % idString)

    # Check keywords
    for entry in newList:
        for name in entry.keywords:
            dbutils.getOrCreateKeyword(name, interactive=interactive)

    # Remove tasks whose lines have been deleted
    for id in oldIds.difference(newIds):
        task = dbutils.getTaskFromId(id)
        session.delete(task)

    # Update existing tasks, add new ones
    nbTasks = len(newList)
    for pos, newEntry in enumerate(newList):
        if newEntry.id:
            task = dbutils.getTaskFromId(newEntry.id)
        else:
            task = Task(creationDate=datetime.now().replace(second=0, microsecond=0), project=project)
            session.add(task)
        task.title = newEntry.title
        task.setKeywordDict(newEntry.keywords)
        task.setStatus(newEntry.status)
        task.urgency = nbTasks - pos


# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Update from version 5 to version 6 of Yokadi DB

@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or newer
"""
from sqlite3 import OperationalError


def createTaskLockTable(cursor):
    cursor.execute("""
create table task_lock (
    id integer not null,
    task_id integer,
    pid integer,
    update_date datetime,
    primary key (id),
    unique (task_id),
    foreign key(task_id) references task (id)
)
""")


def removeTaskTitleUniqConstraint(cursor):
    try:
        cursor.execute("drop index task_uniqTaskTitlePerProject")
    except OperationalError as exc:
        if str(exc) == "no such index: task_uniqTaskTitlePerProject":
            pass


def update(cursor):
    removeTaskTitleUniqConstraint(cursor)
    createTaskLockTable(cursor)


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Update from version 8 to version 9 of Yokadi DB

- Delete invalid TaskKeyword rows
- Move aliases to an Alias table
- Add an uuid column to Task and Project tables

@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or newer
"""
from uuid import uuid1


def deleteInvalidTaskKeywordRows(cursor):
    cursor.execute('delete from task_keyword where task_id is null or keyword_id is null')


def addAliasTable(cursor):
    cursor.execute("""create table alias (
        uuid varchar not null primary key,
        name varchar not null,
        command varchar not null,
        unique(name)
    )""")


def migrateAliases(cursor):
    row = cursor.execute("select value from config where name = 'ALIASES'").fetchone()
    if not row:
        return

    aliasesString = row[0]
    try:
        aliases = eval(aliasesString)
    except Exception:
        # Failed to parse aliases
        print("Failed to parse v9 aliases: {}".format(aliasesString))
        return
    for name, command in aliases.items():
        uuid = str(uuid1())
        cursor.execute("insert into alias(uuid, name, command) values(?, ?, ?)", (uuid, name, command))

    cursor.execute("delete from config where name = 'ALIASES'")


def addUuidColumn(cursor, tableName):
    cursor.execute("alter table {} add column uuid varchar".format(tableName))
    for row in cursor.execute("select id from {}".format(tableName)).fetchall():
        id = row[0]
        uuid = str(uuid1())
        cursor.execute("update {} set uuid = ? where id = ?".format(tableName), (uuid, id))


def update(cursor):
    deleteInvalidTaskKeywordRows(cursor)
    addAliasTable(cursor)
    migrateAliases(cursor)
    addUuidColumn(cursor, "project")
    addUuidColumn(cursor, "task")


if __name__ == "__main__":
    import updateutils
    updateutils.main(update)
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Task test cases
@author: Aurélien Gâteau <mail@agateau.com>
@license: GPL v3 or later
"""
import unittest

from sqlalchemy.orm.exc import NoResultFound

from yokadi.core import dbutils
from yokadi.ycli import tui
from yokadi.ycli.keywordcmd import KeywordCmd, _listKeywords
from yokadi.core.yokadiexception import YokadiException
from yokadi.core import db


class KeywordTestCase(unittest.TestCase):
    def setUp(self):
        db.connectDatabase("", memoryDatabase=True)
        self.session = db.getSession()
        tui.clearInputAnswers()
        self.cmd = KeywordCmd()

    def testKEditNoMerge(self):
        t1 = dbutils.addTask("x", "t1", dict(k1=12, k2=None), interactive=False)
        tui.addInputAnswers("newk1")
        self.cmd.do_k_edit("k1")
        kwDict = t1.getKeywordDict()
        self.assertTrue("k1" not in kwDict)
        self.assertEqual(kwDict.get("newk1"), 12)
        self.assertRaises(YokadiException, dbutils.getKeywordFromName, "k1")

    def testKEditMerge(self):
        t1 = dbutils.addTask("x", "t1", dict(k1=None, k2=None), interactive=False)
        t2 = dbutils.addTask("x", "t2", dict(k1=None), interactive=False)
        tui.addInputAnswers("k2", "y")
        self.cmd.do_k_edit("k1")

        kwDict = t1.getKeywordDict()
        self.assertTrue("k1" not in kwDict)
        self.assertTrue("k2" in kwDict)

        kwDict = t2.getKeywordDict()
        self.assertTrue("k1" not in kwDict)
        self.assertTrue("k2" in kwDict)

        self.assertRaises(YokadiException, dbutils.getKeywordFromName, "k1")

    def testKEditCannotMerge(self):
        """
        One can't merge keywords if they have different values
        """
        t1 = dbutils.addTask("x", "t1", dict(k1=12, k2=None), interactive=False)
        tui.addInputAnswers("k2", "y")
        self.cmd.do_k_edit("k1")
        kwDict = t1.getKeywordDict()
        self.assertTrue("k1" in kwDict)
        self.assertTrue("k2" in kwDict)

        dbutils.getKeywordFromName("k1")

    def testKRemove(self):
        t1 = dbutils.addTask("x", "t1", dict(k1=12, k2=None), interactive=False)
        tui.addInputAnswers("y")
        self.cmd.do_k_remove("k1")
        kwDict = t1.getKeywordDict()
        self.assertFalse("k1" in kwDict)
        self.assertTrue("k2" in kwDict)
        taskKeyword = self.session.query(db.TaskKeyword).filter_by(taskId=t1.id).one()
        self.assertEqual(taskKeyword.keyword.name, "k2")

    def testKRemove_unused(self):
        self.cmd.do_k_add("kw")
        self.session.query(db.Keyword).filter_by(name="kw").one()
        self.cmd.do_k_remove("kw")
        self.assertRaises(NoResultFound, self.session.query(db.Keyword).filter_by(name="kw").one)

    def testKList(self):
        t1 = dbutils.addTask("x", "t1", dict(k1=12, k2=None), interactive=False)
        t2 = dbutils.addTask("x", "t2", dict(k1=None, k3=None), interactive=False)

        lst = list(_listKeywords(self.session))
        lst = [(name, list(ids)) for name, ids in lst]
        self.assertEqual(lst, [("k1", [t1.id, t2.id]),
                               ("k2", [t1.id]),
                               ("k3", [t2.id]),
                               ])
# vi: ts=4 sw=4 et



# -*- coding: UTF-8 -*-
"""
Alias related commands.

@author: Sébastien Renard <Sebastien.Renard@digitalfox.org>
@license: GPL v3 or later
"""
from yokadi.core import db
from yokadi.core.yokadiexception import BadUsageException, YokadiException
from yokadi.ycli.basicparseutils import parseOneWordName
from yokadi.ycli import tui
from yokadi.ycli import colors


class AliasCmd(object):
    def __init__(self):
        self._updateAliasDict()

    def _updateAliasDict(self):
        self.aliases = db.Alias.getAsDict(db.getSession())

    def do_a_list(self, line):
        """List all aliases."""
        if self.aliases:
            lst = sorted(self.aliases.items(), key=lambda x: x[0])
            for name, command in lst:
                print(colors.BOLD + name.ljust(10) + colors.RESET + "=> " + command)
        else:
            print("No alias defined. Use a_add to create one")

    def do_a_add(self, line):
        """Add an alias on a command
        Ex. create an alias 'la' for 't_list -a':
        a_add la t_list -a"""
        tokens = line.split()
        if len(tokens) < 2:
            raise BadUsageException("You should provide an alias name and a command")
        name = tokens[0]
        if name in self.aliases:
            raise YokadiException("There is already an alias named {}.".format(name))
        command = " ".join(tokens[1:])

        session = db.getSession()
        db.Alias.add(session, name, command)
        session.commit()
        self._updateAliasDict()

    def do_a_edit_name(self, line):
        """Edit the name of an alias.
        a_edit_name <alias name>"""
        session = db.getSession()
        name = line
        if name not in self.aliases:
            raise YokadiException("There is no alias named {}".format(name))

        newName = tui.editLine(name)
        newName = parseOneWordName(newName)

        if newName in self.aliases:
            raise YokadiException("There is already an alias named {}.".format(newName))

        session = db.getSession()
        db.Alias.rename(session, name, newName)
        session.commit()
        self._updateAliasDict()

    def do_a_edit_command(self, line):
        """Edit the command of an alias.
        a_edit_command <alias name>"""
        session = db.getSession()
        name = line
        if name not in self.aliases:
            raise YokadiException("There is no alias named {}".format(name))

        command = tui.editLine(self.aliases[name])

        session = db.getSession()
        db.Alias.setCommand(session, name, command)
        session.commit()
        self._updateAliasDict()

    def do_a_remove(self, line):
        """Remove an alias"""
        if line in self.aliases:
            session = db.getSession()
            del self.aliases[line]
            alias = session.query(db.Alias).filter_by(name=line).one()
            session.delete(alias)
            session.commit()
        else:
            tui.error("No alias with that name. Use a_list to display all aliases")


def resolveAlias(line, aliasDict):
    """Look for alias in alias and replace it with rela command
    @param line : string to analyse
    @param aliasDict: aliases dictionnary
    @return: modified line"""
    tokens = line.split()
    if len(tokens) > 0 and tokens[0] in aliasDict:
        line = "%s %s" % (aliasDict[tokens[0]], " ".join(tokens[1:]))
    return line
